/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on April 29, 2010, 7:31 PM
 */

/* For realpath() */
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>

#include <boost/filesystem.hpp>
#include <iostream>
#include <sstream>

#include "u2d_Project.h"
#include "exceptions.h"

namespace bfs = boost::filesystem;


//u2d_Project::u2d_Project() {
//}

u2d_Project::u2d_Project(string project_name) {
    setProjectName(project_name);
}

u2d_Project::~u2d_Project() {
}

void u2d_Project::setProjectName(string project_name) {
    this->project_name = project_name;
}

string u2d_Project::getProjectName() const {
    return project_name;
}

void u2d_Project::setProjectRelativeDir(string project_relative_dir) {
    this->project_relative_dir = project_relative_dir;
}

string u2d_Project::getProjectRelativeDir(bool withName /*= false*/) const {
    if (withName) {
        /* Just as note. project_relative dir is transformed to a path, so then
         the / operator (append) auto promotes the RHS from string to path
         and then we return a string again.*/
        return (bfs::path(project_relative_dir) / this->getProjectName()).string();
    } else {
        return project_relative_dir;
    }
}

void u2d_Project::setProjectAbsoluteDir(string project_absolute_dir) {
    this->project_absolute_dir = project_absolute_dir;
}

string u2d_Project::getProjectAbsoluteDir(bool withName /*= false*/) const {
    if (withName) {
        /* Just as note. project_relative dir is transformed to a path, so then
         the / operator (append) auto promotes the RHS from string to path
         and then we return a string again.*/
        return (bfs::path(project_absolute_dir) / this->getProjectName()).string();
    } else {
        return project_absolute_dir;
    }
}

/**
 * Returns the nodes.projectname file where nodes list
 * is located.
 * If path is true the string returned contains also the path of the
 * file name. If rel is true the path is relative else it's abosulte.
 * If path is false, rel is ignored and the simple filename is returned.
 */
string u2d_Project::getMeshNodesListFilename(bool path /*= true*/, bool rel /*= false*/) const {
    if(path == true && rel == false) {
        return project_absolute_dir + "/nodes." + project_name;
    }
    else if(path == true && rel == true){
        return project_relative_dir + "/nodes." + project_name;
    }
    else {
        return "nodes." + project_name;
    }
}

/**
 * Returns the grid.projectname file where grid connectivity
 * info is located.
 * If path is true the string returned contains also the path of the
 * file name. If rel is true the path is relative else it's absolute.
 * If path is false, rel is ignored and the simple filename is returned.
 * @param path If true the returned string is a complete path else only the
 * filename is returned. Defaults to true.
 * @param rel If true the path returned is relative to where the program was
 * executed else the path is absolute. Defaults to false.
 * @return The requested path.
 */
string u2d_Project::getMeshGridListFilename(bool path /*= true*/, bool rel /*= false*/) const {
    if(path == true && rel == false) {
        return project_absolute_dir + "/grid." + project_name;
    }
    else if(path == true && rel == true){
        return project_relative_dir + "/grid." + project_name;
    }
    else {
        return "grid." + project_name;
    }
}

/**
 * Returns the domain.projectname file where mesh domain
 * info are contained.
 * If path is true the string returned contains also the path of the
 * file name. If rel is true the path is relative else it's abosulte.
 * If path is false, rel is ignored and the simple filename is returned.
 * @param path If true the returned string is a complete path else only the
 * filename is returned. Defaults to true.
 * @param rel If true the path returned is relative to where the program was
 * executed else the path is absolute. Defaults to false.
 * @return The requested path.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
string u2d_Project::getMeshDomainFilename(bool path /*= true*/, bool rel /*= false*/) const {
    if(path == true && rel == false) {
        return project_absolute_dir + "/domain." + project_name;
    }
    else if(path == true && rel == true){
        return project_relative_dir + "/domain." + project_name;
    }
    else {
        return "domain." + project_name;
    }
}

/**
 * Returns the boundary.projectname file where mesh boundary
 * info are contained.
 * If path is true the string returned contains also the path of the
 * file name. If rel is true the path is relative else it's abosulte.
 * If path is false, rel is ignored and the simple filename is returned.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
string u2d_Project::getMeshBoundaryFilename(bool path /*= true*/, bool rel /*= false*/) const {
    if(path == true && rel == false) {
        return project_absolute_dir + "/boundary." + project_name;
    }
    else if(path == true && rel == true){
        return project_relative_dir + "/boundary." + project_name;
    }
    else {
        return "boundary." + project_name;
    }
}

/**
 * Returns the gridplot.projectname file where grid data for plotting is stored
 * If path is true the string returned contains also the path of the
 * file name. If rel is true the path is relative else it's absolute.
 * If path is false, rel is ignored and the simple filename is returned.
 * @param path If true the returned string is a complete path else only the
 * filename is returned. Defaults to true.
 * @param rel If true the path returned is relative to where the program was
 * executed else the path is absolute. Defaults to false.
 * @return The requested path.
 */
string u2d_Project::getMeshGridplotFilename(bool path /*= true*/, bool rel /*= false*/) const {
    if(path == true && rel == false) {
        return project_absolute_dir + "/gridplot." + project_name;
    }
    else if(path == true && rel == true){
        return project_relative_dir + "/gridplot." + project_name;
    }
    else {
        return "gridplot." + project_name;
    }
}

/**
 * Updates the relative path. Call this before any get function that returns a
 * relative path to current if you want to be sure the path is updated relative
 * to _latest_ current path.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Project::updateRelativePath() {
    bfs::path bfspath = bfs::path(this->getProjectAbsoluteDir());
    this->setProjectRelativeDir(NaiveUncomplete(bfspath, bfs::current_path()).string());
}


/**
 * Create a project located in the specified directory. The project name will
 * be deduced from the last part of the given path.
 * i.e.
 * CreateProject("/home/foo/bar") will create a project in /home/foo with
 * project name bar. Hence a directory bar will be created and files will be
 * stored inside it.
 * @param project_path_and_name
 * @return 
 */
u2d_Project *u2d_Project::CreateProject(bfs::path project_path_and_name)
    throw (u2d_IO_Exception &) {
    /* Split it and invoke split constructor */
    return CreateProject(project_path_and_name.parent_path(), project_path_and_name.filename());
}

/**
 * Creates a project object to be used to manage locations of various files.
 * The project will be created in the specified project_path directory following
 * the scheme of naming the project directory as the project itself.
 * i.e.
 * CreateProject("/home/foo", "bar") will create a project in /home/foo. Hence
 * the project directory bar will be created in /home/foo with project files
 * ending up in /home/foo/bar and project name set to bar, so files will be
 * named "bar.something".
 * @param project_path If project path is empty assume the project name refers
 * to the current path. If project path starts with a "/" it is assumed it is
 * and absolute path and project directory will be created there.
 * @param project_name The name to use as project name. A directory with this
 * name will be created in the given project_path.
 * @return 
 */
u2d_Project *u2d_Project::CreateProject(bfs::path project_path, bfs::path project_name)
throw (u2d_IO_Exception &) {
    /* Check project name is not empty */
    if(project_name.empty() || project_name == ".") {
        throw u2d_IO_Exception("Project name empty. Please specify a project name.");
    }

    try {
        bfs::path absolute_project_path = bfs::absolute(bfs::path(project_path / project_name));
        /* Check if the project directory already exists */
        if (bfs::is_directory(absolute_project_path)) {
            throw u2d_IO_Exception("Cannot create project directory, it already exists.");
        } else if (bfs::exists(absolute_project_path)) {
            throw u2d_IO_Exception("Cannot create project directory, a file or link with same name already exists.");
        }
        bfs::create_directory(absolute_project_path);
        /* Normalize/canonize the path */
        absolute_project_path = RealPath(absolute_project_path);

        u2d_Project *project = new u2d_Project(project_name.string());
        project->setProjectAbsoluteDir(absolute_project_path.string());
        project->setProjectRelativeDir(NaiveUncomplete(absolute_project_path, bfs::current_path()).string());
        return project;

    } catch (bfs::filesystem_error &e) {
        std::stringstream msg;
        msg << "Boost::filesystem: " << e.what();
        throw u2d_IO_Exception(msg.str());
    } catch (...) {
        throw;
    }
    
}

/**
 * Load a project given it's full path. The name is assumed to be the last part
 * of the provided path.
 * i.e.
 * LoadProject("/home/foo/bar") will try to load a project in the directory
 * /home/foo/bar with project name bar. Notice that the files must be inside
 * /home/foo/bar and named with the "bar." convention.
 * @param project_path_and_name
 * @return 
 */
u2d_Project *u2d_Project::LoadProject(bfs::path project_path_and_name)
    throw (u2d_IO_Exception &) {   
    /* Split it and invoke split constructor */
    return LoadProject(project_path_and_name, project_path_and_name.filename());
}

/**
 * Load a project in the specified project_path.
 * @param project_path Path where the project _files_ are located.
 * i.e. /home/foo/bar if files are inside bar.
 * @param project_name Name of the project the files are using following the
 * "project_name." convention.
 * @return 
 */
u2d_Project *u2d_Project::LoadProject(bfs::path project_path, bfs::path project_name)
throw (u2d_IO_Exception &) {
    if(project_name.empty() || project_name == ".") {
        throw u2d_IO_Exception("Project name empty. Please specify a project name.");
    }

     try {
        /* Transform the project path to absolute */
        bfs::path absolute_project_path = bfs::absolute(project_path);
        /* Check if the project directory already exists */
        if (!bfs::exists(absolute_project_path)) {
            throw u2d_IO_Exception("Cannot load project. Project is inexistent.");
        } else if (!bfs::is_directory(absolute_project_path)) {
            throw u2d_IO_Exception("Cannot load project, path provided is not a directory.");
        }
        /* Normalize/canonize the path */
        absolute_project_path = RealPath(absolute_project_path);

        /* Create the project and set it's paths */
        u2d_Project *project = new u2d_Project(project_name.string());
        project->setProjectAbsoluteDir(absolute_project_path.string());
        project->setProjectRelativeDir(NaiveUncomplete(absolute_project_path, bfs::current_path()).string());
        return project;

    } catch (bfs::filesystem_error &e) {
        std::stringstream msg;
        msg << "Boost::filesystem: " << e.what();
        throw u2d_IO_Exception(msg.str());
    } catch (...) {
        throw;
    }
}

void u2d_Project::dump() {
    std::cout << "--== Project Paths: ==--" << std::endl;
    std::cout << "Project Name: " << getProjectName() << std::endl;
    std::cout << "Project Absolute Path: " << getProjectAbsoluteDir() << std::endl;
    std::cout << "Project Relative Path: " << getProjectRelativeDir() << std::endl;
    std::cout << "Project Absolute Path with project Name: " << getProjectAbsoluteDir(true) << std::endl;
    std::cout << "Project Relative Path with project Name: " << getProjectRelativeDir(true) << std::endl;
    std::cout << "--== Absolute paths: ==--" << std::endl;
    std::cout << "Nodes path: " << getMeshNodesListFilename() << std::endl;
    std::cout << "Grid path: " << getMeshGridListFilename() << std::endl;
    std::cout << "Domain path: " << getMeshDomainFilename() << std::endl;
    std::cout << "Boundary path: " << getMeshBoundaryFilename() << std::endl;
    std::cout << "--==  Relative paths: ==--" << std::endl;
    std::cout << "Nodes path: " << getMeshNodesListFilename(true, true) << std::endl;
    std::cout << "Grid path: " << getMeshGridListFilename(true, true) << std::endl;
    std::cout << "Domain path: " << getMeshDomainFilename(true, true) << std::endl;
    std::cout << "Boundary path: " << getMeshBoundaryFilename(true, true) << std::endl;
}

/* A naive uncomplete function for boost path */
bfs::path u2d_Project::NaiveUncomplete(bfs::path const path,
        bfs::path const base) {
    if (path.has_root_path()){
        if (path.root_path() != base.root_path()) {
            return path;
        } else {
            return NaiveUncomplete(path.relative_path(), base.relative_path());
        }
    } else {
        if (base.has_root_path()) {
            throw "cannot uncomplete a path relative path from a rooted base";
        } else {
            typedef bfs::path::const_iterator path_iterator;
            path_iterator path_it = path.begin();
            path_iterator base_it = base.begin();
            while ( path_it != path.end() && base_it != base.end() ) {
                if (*path_it != *base_it) break;
                ++path_it; ++base_it;
            }
            bfs::path result;
            for (; base_it != base.end(); ++base_it) {
                result /= "..";
            }
            for (; path_it != path.end(); ++path_it) {
                result /= *path_it;
            }
            return result;
        }
    }
}

/**
 * Wrap the POSIX realpath() function that canonize/normalize/resolve
 * a path following simlinks, ".." and ".", if we are not POSIX, an alternative
 * and potentially less reliable (but more portable) boost::filesystem based
 * implementation is provided.
 * 
 * @param path The path to resolve. It is not modified.
 * @return A boost::filesystem path pointing to the same location as the given
 * path but normalized/resolved.
 */
bfs::path u2d_Project::RealPath(bfs::path const originalPath) {
#if  HAVE_REALPATH
    char *resolvedPathChar = realpath(originalPath.c_str(), NULL);
    /* In case of failure return the original path */
    if(resolvedPathChar == NULL) {
        return originalPath;
    }
    bfs::path resolvedPath(resolvedPathChar);
    free(resolvedPathChar);
    return resolvedPath;
#else
    bfs::path result;
    for (bfs::path::iterator it = originalPath.begin();
            it != originalPath.end();
            ++it) {
        if (*it == "..") {
            // /a/b/.. is not necessarily /a if b is a symbolic link
            if (bfs::is_symlink(result))
                result /= *it;
                // /a/b/../.. is not /a/b/.. under most circumstances
                // We can end up with ..s in our result because of symbolic links
            else if (result.filename() == "..")
                result /= *it;
                // Otherwise it should be safe to resolve the parent
            else
                result = result.parent_path();
        } else if (*it == ".") {
            // Ignore
        } else {
            // Just cat other path entries
            result /= *it;
        }
    }
    return result;  
#endif

}
