/* 

    umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2007 Gabriele Dini Ciacci

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef U2D_DISPLAY_SET_H
#define U2D_DISPLAY_SET_H

#include <vector>
#include <string>

#include <GL/gl.h>

#include "common/exceptions.h"
#include "common/u2d_Mesh.h"
#include "solver/u2d_Problem.h"
#include "gl/u2d_DisplayObject.h"
#include "gl/u2d_DisplayContainer.h"

#define VERTEX_PRECISION GLfloat
#define VERTEX_PRECISION_TYPE GL_FLOAT

class u2d_DisplaySet : public u2d_DisplayObject {
public:
    u2d_DisplaySet(std::string _name = "") throw (u2d_OpenGL_Exception &);
    ~u2d_DisplaySet();
    
    void draw();
    
    /* Named parameters chains */
    u2d_DisplaySet& enable();
    u2d_DisplaySet& disable();
    u2d_DisplaySet& setFillMode(GLenum _fillMode);
    u2d_DisplaySet& setLineWidth(GLfloat _lineWidth);
    u2d_DisplaySet& setLineColor(GLfloat _red, GLfloat _green, GLfloat _blue);
    u2d_DisplaySet& setName(std::string _name);
    std::string getName() const;

    /* States */
    bool isEnabled() const;
        
    /* Vertex operations */
    u2d_DisplaySet& setCoordPerVertex(GLint m_coord_per_vertex);
    GLint getCoordPerVertex() const;
    void pushVertexCoord(VERTEX_PRECISION a);
    void pushVertexCoord(VERTEX_PRECISION a, VERTEX_PRECISION b);
    void pushVertexCoord(VERTEX_PRECISION a, VERTEX_PRECISION b, VERTEX_PRECISION c);
    GLsizei getVertexCount() const;
    GLsizei getVertexCoordCount() const;
    VERTEX_PRECISION *getVertexArray();

    /* Topology operations */
    u2d_DisplaySet& setVertexPerElement(GLint m_coord_per_topology);
    GLint getVertexPerElement() const;
    void pushTopologyElement(GLuint a);
    void pushTopologyElement(GLuint a, GLuint b);
    void pushTopologyElement(GLuint a, GLuint b, GLuint c);
    GLsizei getElementCount() const;
    GLsizei getTopologyElementCount() const;
    GLuint *getTopologyArray();
    
    /* Color operations */
    u2d_DisplaySet& setColorsPerVertex(GLint _colorsPerVertex) throw (u2d_OpenGL_Exception &);
    GLint getColorsPerVertex() const;
    void pushVertexColor(GLfloat a, GLfloat b, GLfloat c);
    GLsizei getColorCount() const;
    GLsizei getColorValuesCount() const;
    GLfloat *getColorArray();

    /* Structure of the given data */
    u2d_DisplaySet& setVertexType(GLenum _vertexType) throw (u2d_OpenGL_Exception &);
    GLenum getVertexType() const;
       
    /* Static converters */
    static u2d_DisplaySet *FromMeshDomain(u2d_Mesh *mesh);
    static u2d_DisplayContainer *FromMeshDomainWithSolution(u2d_Mesh *mesh, u2d_Problem *problem);
    static u2d_DisplaySet *FromMeshBoundary(u2d_Mesh *mesh);
    static u2d_DisplayContainer *CreateKey(u2d_Mesh* mesh, u2d_Problem *problem);
    static u2d_DisplaySet *CreateKeyGradient(u2d_Mesh* mesh, u2d_Problem *problem);

private:
    void destroyDL();
    void createDL() throw (u2d_OpenGL_Exception &);
    void destroyVBO();
    void createVBO() throw (u2d_OpenGL_Exception &);
    
    /* Compile functions */
    void compileDL() throw (u2d_OpenGL_Exception &);
    void compileVBO() throw (u2d_OpenGL_Exception &);
    
    /* Implementation specific functions */
    void drawVA();
    void drawDL();
    void drawVBO();
    
    /* Structure of the given data */
    GLenum vertexType;
        
    /* Set name*/
    string name;
    
    /* States */
    bool compiledDL;
    bool compiledVBO;
    bool drawing;
    bool enabled;
    
    /* Drawing options*/
    GLenum fillMode;
    GLfloat lineWidth;
    GLfloat lineColor[3];

    /* Data structure information */
    GLint coordPerVertex;
    GLint vertexPerElement;
    GLint colorPerVertex;
    
    /* For display list */
    GLuint displayList;
    
    /* For Video Buffer Object implementation */
    GLuint VBOVertexBuffer;
    GLuint VBOTopologyBuffer;

    /* The data */
    std::vector<VERTEX_PRECISION> vertexArray;
    std::vector<GLuint> topologyArray;
    std::vector<GLfloat> colorArray;
    
    void dump() const;
};


#endif
