/* 
 * File:   IsoElement2DP13Points.cpp
 * Author: Gabriele Dini Ciacci <dark.schneider@iol.it>
 * 
 * Created on January 17, 2010, 8:15 PM
 */

#include <math.h>

#include "u2d_IsoElement2DP13Points.h"
#include "u2d_Mesh.h"
#include "u2d_GaussPoints.h"


using namespace mtl;

/**
 * This represents and Isometric Element for 2D problems
 * with linear interpolation and 3 gauss points. It's on
 * a triangular element!
 * We are linear, so we have same number of degree of freedom
 * as we have nodes per element.
 */
u2d_IsoElement2DP13Points::u2d_IsoElement2DP13Points() :
gauss_points_n(3),
gauss_points(2, gauss_points_n),
w(P1_2D_DEGREE_FREEDOM, gauss_points_n),
dx(P1_2D_DEGREE_FREEDOM, gauss_points_n),
dy(P1_2D_DEGREE_FREEDOM, gauss_points_n),
p(gauss_points_n) {

    gauss_points(0, 0) = 1.0 / 6.0;
    gauss_points(1, 0) = 1.0 / 6.0;

    gauss_points(0, 1) = 2.0 / 3.0;
    gauss_points(1, 1) = 1.0 / 6.0;

    gauss_points(0, 2) = 1.0 / 6.0;
    gauss_points(1, 2) = 2.0 / 3.0;

    /* Do this for all gauss points couples (x,y)
     * that we have just defined above */
    for (int i = 0; i < gauss_points_n; i++) {
        /* For the first vertex of the simplex, evaluate it's
         * corresponding shape function at the gauss point.
         * The evaluate the derivative, in x and y
         */
        w(0, i) = shapeP1Node1(gauss_points(0, i), gauss_points(1, i));
        dx(0, i) = -1.0;
        dy(0, i) = -1.0;

        /* For the second vertex of the simplex, evaluate it's
         * corresponding shape function at the gauss point
         * The evaluate the derivative, in x and y
         */
        w(1, i) = shapeP1Node2(gauss_points(0, i), gauss_points(1, i));
        dx(1, i) = 1.0;
        dy(1, i) = 0.0;

        /* For the third vertex of the simplex, evaluate it's
         * corresponding shape function at the gauss point
         * The evaluate the derivative, in x and y
         */
        w(2, i) = shapeP1Node3(gauss_points(0, i), gauss_points(1, i));
        dx(2, i) = 0.0;
        dy(2, i) = 1.0;

        p(i) = 1.0 / 6.0;
    }

}

u2d_IsoElement2DP13Points::u2d_IsoElement2DP13Points(const u2d_IsoElement2DP13Points& orig) {
}

u2d_IsoElement2DP13Points::~u2d_IsoElement2DP13Points() {
}

/**
 * Evaluate and store the values of derivatives and of the
 * jacobian determinant at Gauss points of all domain elements.
 */
u2d_DomainGaussPoints *u2d_IsoElement2DP13Points::applyTo(u2d_Mesh *mesh) {
    u2d_DomainGaussPoints *dgp;
    dgp = new u2d_DomainGaussPoints(mesh->getDomainElementNumber());
    // evaluate and store the values of derivatives and of the
    // jacobian determinant at Gauss points of all volume elements

    /* For the linear element the derivatives are constant and
     * the weights are uniform so the first Gauss point is used
     */
    int gauss_point = 0; /* The first */

    /* We calculate dd_x and dd_y vectors for the first gauss
     * point outside the element loop, because they are the
     * same for all the elements, since they depend on the isoelement only
     */
    dense_vector<double> dd_x(DOMAIN_ELEMENT_SIZE);
    dense_vector<double> dd_y(DOMAIN_ELEMENT_SIZE);
    for (int ds = 0; ds < DOMAIN_ELEMENT_SIZE; ds++) {
        dd_x(ds) = dx(ds, gauss_point);
        dd_y(ds) = dy(ds, gauss_point);
    }

    for (int i = 0; i < mesh->getDomainElementNumber(); i++) {

        /* OLD VERSION
        dense_vector<double> rr_x(DOMAIN_ELEMENT_SIZE);
        dense_vector<double> rr_y(DOMAIN_ELEMENT_SIZE);
        for (int ds; ds++; ds < DOMAIN_ELEMENT_SIZE) {
            rr_x(ds) = mesh->rr(1, mesh->jjd(ds, i) - 1);
            rr_y(ds) = mesh->rr(2, mesh->jjd(ds, i) - 1);
        }
        dense2D<double> dr(2, 2);
        dr(0, 0) = dot(rr_x, dd_x);
        dr(0, 1) = dot(rr_x, dd_y);
        dr(1, 0) = dot(rr_y, dd_x);
        dr(1, 1) = dot(rr_y, dd_y);
         */

        /* New version */
        multi_vector<dense_vector<double> > elem_i = mesh->getDomainElementNodes(i + 1);
        /* Assemble the jacobian */
        dense2D<double> dr(2, 2);
        dr(0, 0) = dot(elem_i.vector(0), dd_x);
        dr(0, 1) = dot(elem_i.vector(0), dd_y);
        dr(1, 0) = dot(elem_i.vector(1), dd_x);
        dr(1, 1) = dot(elem_i.vector(1), dd_y);
        /* Calculate the jacobian determinant */
        double jacobian = dr(0, 0) * dr(1, 1) - dr(0, 1) * dr(1, 0);


        for (int n = 0; n < DOMAIN_ELEMENT_SIZE; n++) {
            dgp->dw_x(n, i) = (dd_x(n) * dr(1, 1) - dd_y(n) * dr(1, 0)) / jacobian;
            dgp->dw_y(n, i) = (-dd_x(n) * dr(0, 1) + dd_y(n) * dr(0, 0)) / jacobian;
        }
        dgp->jacobian_weight(i) = p(gauss_point) * jacobian;
    }
    // Lovely debug output
    //std::cout << dw_x << std::endl;
    //std::cout << dw_y << std::endl;
    //std::cout << jacobian_weight << std::endl;

    return dgp;
}