/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on January 24, 2010, 6:56 PM
 */

#include <algorithm> // for find()

#include "u2d_NeumannCondition.h"

using namespace std;

u2d_NeumannCondition::u2d_NeumannCondition() :
neumann_element_n(0),
neumann_elements(0) {

}

u2d_NeumannCondition::u2d_NeumannCondition(const u2d_NeumannCondition& orig) {
}

u2d_NeumannCondition::~u2d_NeumannCondition() {
}

/**
 * Count all nodes of the boundary elements which belong to
 * the sides where a Neumann boundary condition is prescribed.
 */
void u2d_NeumannCondition::countElements(u2d_Mesh* mesh) {

    neumann_element_n = 0;
    for (int i = 0; i < mesh->getBoundaryElementNumber(); i++) {
        if (isNeumann(mesh->getElementSideId(i + 1))) {
            ++neumann_element_n;
        }
    }
    /* Resize the vector accordingly */
    neumann_elements.change_dim(neumann_element_n);
    // Lovely debug output
    //cout << neumann_element_n << endl;

}

/**
 * Generation of the list of surface elements for
 * enforcing NONhomogeneous Neumann conditions in weak
 * form through the surface integral.
 */
void u2d_NeumannCondition::calculateElements(u2d_Mesh* mesh) {
    countElements(mesh);
    int cur_elem = 0;
    for (int i = 0; i < mesh->getBoundaryElementNumber(); i++) {
        if (isNeumann(mesh->getElementSideId(i + 1))) {
            neumann_elements(cur_elem) = i + 1;
            cur_elem++;
        }
    }
    // Lovely debug output
    //cout << jjb_neu << endl;
}

/**
 *  Returns the number of Neumann elements.
 */
size_t u2d_NeumannCondition::getNuemannElementNumber() {
    return neumann_element_n;
}

/**
 * Set the given side id as a side where to enforce
 * the NeumannCondition.
 * @return The current total number of sides that have been set as Neumann
 * condition sides.
 * i.e. The total number of sides where the Neumann boundary condition
 * is to be enforced:
 * NeumannCondition nc;
 * nc.setside(4) // return 1
 * nc.setside(1) // return 2
 * nc.setside(5) // return 3
 */
int u2d_NeumannCondition::setSide(int side_id) {
    neu_sides.push_back(side_id);
    return neu_sides.size();
}

/**
 * Say if a given side id is in the list of sides where to enforce
 * the Neumann boundary condition.
 * @return Return true if the side_id is in the list and Neumann condition
 * must be enforced on it. Return false otherwise.
 */
bool u2d_NeumannCondition::isNeumann(int side_id) {
    if (find(neu_sides.begin(), neu_sides.end(), side_id) != neu_sides.end()) {
        return true;
    }
    return false;
}
