/*

    umesh2dgui, a GUI for umesh2d grid generator

    Copyright (C) 2010-2011 Gabriele Dini Ciacci <dark.schneider@iol.it>
    Based on a code from Luigi Quartapelle and Massimo Biava

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Created on March 8, 2010, 7:25 PM
 */


#include "u2d_Helmholtz.h"

/* For solver */
#include "u2d_Mesh.h"
#include "u2d_load_umesh2d.h"
#include "u2d_IsoElement2DP13Points.h"
#include "u2d_IsoElement1DP12Points.h"
#include "u2d_DirichletCondition.h"
#include "u2d_NeumannCondition.h"
#include "u2d_GaussPoints.h"
#include "u2d_Problem.h"
#include "u2d_ProblemCalculator.h"
#include "u2d_Constants.h"
#include "u2d_Solver.h"
#include "solver/u2d_Constants.h"
#include "solver/u2d_Solver.h"
#include "file_output/u2d_PlotmtvOutput.h"
#include "file_output/u2d_ScalarFieldOutput.h"
#include "file_output/u2d_VectorFieldOutput.h"
#include "solver/u2d_NeumannCondition.h"
#include "u2d_ScalarFieldOutput.h"
#include "common/u2d_Math.h"
#include "common/u2d_Project.h"

#include "loaders/u2d_load_scalarfield.h"
#include "loaders/u2d_load_vectorfield.h"



/**
 * Creates an Helmholtz problem given a u2d_Project.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz(u2d_Project *project) :
alpha(0),
gamma(0),
//source(mesh->getDomainNodeNumber(), 0.0),
source(0, 0.0),
avector(0, 0.0),
bvector(0, 0.0),
//kvector(mesh->getDomainNodeNumber(), 0.0),
kvector(0, 0.0),
bfield(0, 0),
//vfield(mesh->getSpatialDimension(), mesh->getDomainNodeNumber())
vfield(0, 0) {
    u2d_Load_Umesh2d mesh_loader(project->getProjectAbsoluteDir(), project->getProjectName(), true);
    this->mesh = mesh_loader.getMesh();
    //this->vfield = 0.0;
    //this->bfield = 0.0;
}

/**
 * Creates an Helmholtz problem give path and the project name where
 * the mesh files are located. The files must be in umesh2d format named
 * grid.problemname and nodes.problemname that is the standard naming for
 * files created with umesh2d.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz(string meshdir, string problemname) :
alpha(0),
gamma(0),
//source(mesh->getDomainNodeNumber(), 0.0),
source(0, 0.0),
avector(0, 0.0),
bvector(0, 0.0),
//kvector(mesh->getDomainNodeNumber(), 0.0),
kvector(0, 0.0),
bfield(0, 0),
//vfield(mesh->getSpatialDimension(), mesh->getDomainNodeNumber())
vfield(0, 0) {
    u2d_Load_Umesh2d mesh_loader(meshdir, problemname, true);
    this->mesh = mesh_loader.getMesh();
    //this->vfield = 0.0;
    //this->bfield = 0.0;
}

/**
 * Creates an Helmholtz problem given a pointer to an already loaded mesh.
 * TODO: The mesh is copied internally and so the passed pointer can be
 * deallocated. Currently the mesh is not copied, but the destructor will try
 * to deallocate it.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Helmholtz::u2d_Helmholtz(u2d_Mesh *_mesh) :
alpha(0),
gamma(0),
//source(mesh->getDomainNodeNumber(), 0.0),
source(0, 0.0),
avector(0, 0.0),
bvector(0, 0.0),
//kvector(mesh->getDomainNodeNumber(), 0.0),
kvector(0, 0.0),
bfield(0, 0),
//vfield(mesh->getSpatialDimension(), mesh->getDomainNodeNumber())
vfield(0, 0) {
    mesh = _mesh;
    // We should really do a deep copy of the mesh object and not use
    // The external one. So something like:
    //    mesh = new u2d_Mesh(*_mesh);
    // But that constructor is not implemented yet.
    //this->vfield = 0.0;
    //this->bfield = 0.0;
}

u2d_Helmholtz::u2d_Helmholtz(const u2d_Helmholtz& /*orig*/) :
alpha(0),
gamma(0),
//source(mesh->getDomainNodeNumber(), 0.0),
source(0, 0.0),
avector(0, 0.0),
bvector(0, 0.0),
//kvector(mesh->getDomainNodeNumber(), 0.0),
kvector(0, 0.0),
bfield(0, 0),
//vfield(mesh->getSpatialDimension(), mesh->getDomainNodeNumber())
vfield(0, 0) {
}

u2d_Helmholtz::~u2d_Helmholtz() {
    delete mesh;
}

/**
 * Solve the Helmholtz problem, that is:
 *
 *  [ - D.(k(r)D )  +  v(r).D  +  gamma ] u  =  source(r)
 *
 * \nabla \cdot k \left( r \right) \nabla u + v \left( r \right) \nabla u + \gamma u = s \left( r \right)
 *
 * The boundary conditions can be both Dirichlet and Robin, that is:
 *
 *    |
 *  u |    =  a(s)
 *    |S_D
 *
 * u |_{\partial \Omega} = a \left( s \right)
 *
 *     /                  du  \ |
 *    | alpha u  +  k(r)----   ||     =  b(s)
 *     \                  dn  / |S_R
 *
 * \left( \alpha u + k \left( r \right) \frac{\mathrm{d}u}{\mathrm{d}n} \right) = b \left( s \right)
 *
 * Alpha is called alpha, and determines if the boundary datum is of Neumann or
 * Robin type.
 * Gamma is called gamma.
 * K(r) is called thermal conductivity.
 * V(r) is called advection.
 * Source(r) is called source.
 * 
 * This function expects that the mesh is already loaded. The constructor
 * will load the mesh for you.
 * @return A u2d_Problem containing the A Matrix, the RHS b term and the solution
 * x, like Ax = b.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
u2d_Problem *u2d_Helmholtz::run() {
    /* Generate Gauss points and boundary condition points */
    dgp = ie2d.applyTo(mesh);
    bgp = ie1d.applyTo(mesh);
    dc.generateNodes(mesh);
    nc.calculateElements(mesh);
    
    /* If some vectors has not been calculated from formulas do it now,
     * after the boundary condition is ready */
    this->calculateVectors();

    /* Create a problem and a calculator on that problem for terms */
    problem = new u2d_LinearProblem<dense2D<double> >(mesh->getDomainNodeNumber());
    calculator = new u2d_ProblemCalculator(problem, mesh);
    std::cout << "--== Calculating problem matrix and RHS ==--" << std::endl;
    /* LHS (Matrix) terms */
    calculator->addDW__AlphaDG_left(dgp, kvector);
    calculator->addW__AlphaDG_left(dgp, vfield);
    calculator->addAlphaW__G_left(dgp, gamma);
    calculator->addAlphaW__G_surf_left(bgp, &nc, alpha);
    calculator->forceDirichlet_left(&dc);
    /* RHS terms */
    calculator->addW__G_right(dgp, source);
    calculator->addW__G_surf_right(bgp, &nc, bvector);
    calculator->addW__divG_surf_right(bgp, &nc, bfield);
    calculator->forceDirichlet_right(&dc, avector);
    std::cout << "--== Done ==--" << std::endl;
    
    /* A solutor... and solve :) */
    std::cout << "--== Please wait, solving problem... ==--" << std::endl;
    u2d_LinearSolver<dense2D<double> > solver(problem);
    solver.solve();
    std::cout << "--== Done ==--" << std::endl;
    
    return problem;
}

void u2d_Helmholtz::setSideNeumann(int side_id) {
    nc.setSide(side_id);
}

void u2d_Helmholtz::setSideDirichlet(int side_id) {
    dc.setSide(side_id);
}

void u2d_Helmholtz::setGamma(double _gamma) {
    gamma = _gamma;
}

void u2d_Helmholtz::setAlpha(double _alpha) {
    alpha = _alpha;
}

/**
 * Set the vector of Dirichlet condition.
 * Set the vector a(r) that is the Dirichlet condition on the Dirichlet boundary,
 * consequently this must have length equal to the number of Dirichlet nodes of
 * the boundary. The nodes are numbered by boundary numeration and ordered as
 * the unique node set that is generated by the u2d_DirichletCondition class.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setAVector(dense_vector<double> _avector) {
    /* check that _avector is the same size as Dirichlet condition sides */
    /* eventually resize avector to the appropriate size */
    this->avector = mtl::clone(_avector);
}

/**
 * Set the vector for the Dirichlet condition.
 * Set the vector a(r) that is the Dirichlet condition on the Dirichlet boundary,
 * consequently this must have length equal to the number of Dirichlet nodes of
 * the boundary. The nodes are numbered by boundary numeration and ordered as
 * the unique node set that is generated by the u2d_DirichletCondition class.
 * @param equation The equation to evaluate on the Dirichlet boundary.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it> 
 */
void u2d_Helmholtz::setAVector(string equation) {
    this->avector_equation = equation;
}

/**
 * Set the vector of Robin condition.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBVector(dense_vector<double> _bvector) {
    /* add exception if wrong size size of boundary nodes */
    this->bvector = mtl::clone(_bvector);
}

/**
 * Set the vector of Robin condition.
 * This will evaluated on _all_ boundary nodes and not only on nodes that
 * belong to Robin elements. Naturally values on non Robin belonging elements
 * will be ignored.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBVector(string equation) {
    /* add exception if wrong size size of boundary nodes */
    this->bvector_equation = equation;
}

/**
 * Set the Thermal (K) Conductivity scalar field.
 * The field is represented as a vector where each element is the value of the
 * thermal conductivity in the corresponding mesh node. Consequently this vector
 * has length equal to the number of domain nodes.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setKVector(dense_vector<double> _kvector) {
    this->kvector = mtl::clone(_kvector);
}

/**
 * Set the Thermal (K) Conductivity scalar field.
 * The field is represented as a vector where each element is the value of the
 * thermal conductivity in the corresponding mesh node. Consequently this vector
 * has length equal to the number of domain nodes.
 * This will be evaluated on all nodes of the domain.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setKVector(string equation) {
    this->kvector_equation = equation;
}

/**
 * Set the vector field of Robin condition.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimensions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 * 
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBField(dense2D<double> _bfield) {
    /* add exception if wrong size ????size of boundary nodes */
    this->bfield = mtl::clone(_bfield);
}

/**
 * Set the vector field of Robin condition.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimensions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 * 
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setBField(string equationx, string equationy) {
    this->bfieldx_equation = equationx;
    this->bfieldy_equation = equationy;
}

/**
 * Set the Advection velocity vector field.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimensions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 *
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setVField(dense2D<double> _vfield) {
    /* add exception if wrong size */
    this->vfield = mtl::clone(_vfield);
}

/**
 * Set the Advection velocity vector field.
 * The filed is represented as N row vectors one above the other. That is:
 * It is a NxM matrix where N is the number of spatial dimensions the problem
 * has and M is the number of domain nodes.
 * I.e. in 2D:
 *
 *   V = | 0.345   0.762   ...   0.997 |   value of X
 *       | 0.234   0.748   ...   0.134 |   value of Y
 *         ^^^^^   ^^^^^         ^^^^^
 * Domain Node 1   D.Node 2      D.Node M
 *
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setVField(string equationx, string equationy) {
    /* add exception if wrong size */
    this->vfieldx_equation = equationx;
    this->vfieldy_equation = equationy;
}

/**
 * Sets the source scalar field.
 * The field is represented as a vector where each element is the intensity of
 * the source in the corresponding mesh node. Consequently this vector
 * has length equal to the number of domain nodes.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setSource(dense_vector<double> _source) {
    /* add exception if wrong size */
    this->source = mtl::clone(_source);
}

/**
 * Sets the source scalar field.
 * The field is represented as a vector where each element is the intensity of
 * the source in the corresponding mesh node. Consequently this vector
 * has length equal to the number of domain nodes.
 * The passed equation is evaluated in each point of the domain as per use
 * of u2d_Math::ScalarEvaluator().
 * @param equation The equation to evaluate in each mesh point.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::setSource(string equation) {
    this->source_equation = equation;
}

/**
 * Gets a matrix with the coordinates of the nodes that belong to Dirichlet side.
 * @return A matrix where each column contains the coordinates of the
 * corresponding boundary element. The row length is the number of
 * Dirichlet boundary elements.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
dense2D<double> u2d_Helmholtz::getDirichletNodes() {
    /* Build a matrix with coordinates of boundary Dirichlet nodes */
    dense2D<double> rr_dirichlet(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    std::set<int>::iterator it;
    int rr_d_index = 0;
    for (it = dc.dirichlet_unique_nodes.begin();
            it != dc.dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rr_dirichlet(j, rr_d_index) = mesh->rr(j, *it - 1);
        }
        ++rr_d_index;
    }
    return rr_dirichlet;
}

/**
 * 
 * @return A matrix where each column contains the coordinates of the
 * corresponding boundary element. The row length is the number of
 * all boundary elements.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
dense2D<double> u2d_Helmholtz::getBoundaryNodes() {
    dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
        }
    }
    return rrb;
}

/**
 * Calculates the vectors that have been given with formulas and not set
 * directly.
 * This function must be called after the Neumann and Dirichlet conditions has
 * been calculated else they will be unable to determine the vectors because
 * the boundary points on which evaluation is needed are not yet known. i.e.
 * we do not know if a boundary point is Dirichlet or Robin.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::calculateVectors() {
    if (this->source_equation.size() > 0) {
        this->source = u2d_Math::ScalarEvaluator(this->mesh->rr, this->source_equation);
    }
    if (this->avector_equation.size() > 0) {
        this->avector = u2d_Math::ScalarEvaluator(this->getDirichletNodes(), this->avector_equation);
    }
    if (this->bvector_equation.size() > 0) {
        this->bvector = u2d_Math::ScalarEvaluator(this->getBoundaryNodes(), this->bvector_equation);
    }
    if (this->bfieldx_equation.size() > 0 || this->bfieldy_equation.size() > 0) {
        /* Set the two equations for calculating the field. If one has been
         left empty, then set it to zero. Anyway setter methods should prevent
         this occurrence. */
        this->bfield = 
                u2d_Math::FieldEvaluator(this->getBoundaryNodes(),
                this->bfieldx_equation.size() > 0 ? this->bfieldx_equation : string("0"),
                this->bfieldy_equation.size() > 0 ? this->bfieldy_equation : string("0"));
    }
    if (this->kvector_equation.size() > 0) {
        this->kvector = u2d_Math::ScalarEvaluator(this->mesh->rr, this->kvector_equation);
    }
    if (this->vfieldx_equation.size() > 0 || this->vfieldy_equation.size() > 0) {
        /* Set the two equations for calculating the field. If one has been
         left empty, then set it to zero. Anyway setter methods should prevent
         this occurrence. */
        this->vfield = 
                u2d_Math::FieldEvaluator(this->mesh->rr,
                this->vfieldx_equation.size() > 0 ? this->vfieldx_equation : string("0"),
                this->vfieldy_equation.size() > 0 ? this->vfieldy_equation : string("0"));
    }
}


/**
 * Run a self contained test for a Helmholtz problem on a square with known
 * analytical solution. The Difference from exact solution is printed on screen.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::SquareTest() {
    /* Load the mesh */
    u2d_Load_Umesh2d mesh_loader("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square", "square_c", true);
    u2d_Mesh *mesh = mesh_loader.getMesh();
    /* Calculate the iso elements and the domain and boundary gauss points */
    u2d_IsoElement2DP13Points ie2d;
    u2d_IsoElement1DP12Points ie1d;
    u2d_DomainGaussPoints *dgp = ie2d.applyTo(mesh);
    u2d_BoundaryGaussPoints *bgp = ie1d.applyTo(mesh);
    /* Calculate Dirichlet Boundary Condition */
    u2d_DirichletCondition dc;
    dc.setSide(1);
    dc.setSide(2);
    dc.setSide(3);
    dc.generateNodes(mesh);
    /* Calculate Von Neumann Condition */
    u2d_NeumannCondition nc;
    nc.setSide(4);
    nc.calculateElements(mesh);


    /* Create a problem */
    u2d_LinearProblem<dense2D<double> > problem(mesh->getDomainNodeNumber());
    /* Calculate various terms of the problem */
    u2d_ProblemCalculator calculator(&problem, mesh);
    
    /* LHS (Matrix) terms */
    calculator.addDW__AlphaDG_left(dgp, u2d_Constants::thermalConductivity(mesh->rr));
    calculator.addW__AlphaDG_left(dgp, u2d_Constants::advection(mesh->rr));
    double gamma = 10.0;
    calculator.addAlphaW__G_left(dgp, gamma);
    double alpha = 2.0;
    calculator.addAlphaW__G_surf_left(bgp, &nc, alpha);
    calculator.forceDirichlet_left(&dc);
    
    /* RHS terms */
    calculator.addW__G_right(dgp, u2d_Constants::source(mesh->rr, gamma));
    /* Build a matrix with coordinates of boundary nodes */
    dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
        }
    }
    /* Prepare the solution for addW__G_right on boundary */
    dense_vector<double> pyt_alpha(mesh->getBoundaryNodeNumber());
    pyt_alpha = alpha * u2d_Constants::pythagoras(rrb);
    calculator.addW__G_surf_right(bgp, &nc, pyt_alpha);
    /* Prepare the field for addW__divG_right on the boundary */
    dense_vector<double> thermal_k(num_cols(rrb));
    thermal_k = u2d_Constants::thermalConductivity(rrb);
    dense2D<double> spread_thermal_k(2, num_rows(thermal_k));
    spread_thermal_k[0][mtl::irange()] = trans(thermal_k);
    spread_thermal_k[1][mtl::irange()] = trans(thermal_k);
    dense2D<double> grad_pyt(mesh->getSpatialDimension(), num_cols(rrb));
    grad_pyt = u2d_Constants::gradPythagoras(rrb);
    dense2D<double> field(mesh->getSpatialDimension(), num_cols(rrb));
    field = ele_prod(spread_thermal_k, grad_pyt);
    calculator.addW__divG_surf_right(bgp, &nc, field);
    /* Force the boundary condition on the RHS terms */
    /* Build a matrix with coordinates of boundary Dirichlet nodes */
    dense2D<double> rr_dirichlet(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    std::set<int>::iterator it;
    int rr_d_index = 0;
    for (it = dc.dirichlet_unique_nodes.begin();
            it != dc.dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rr_dirichlet(j, rr_d_index) = mesh->rr(j, *it - 1);
        }
        ++rr_d_index;
    }
    calculator.forceDirichlet_right(&dc, u2d_Constants::pythagoras(rr_dirichlet));
    /* Let's try a solutor */
    u2d_LinearSolver<dense2D<double> > solver(&problem);
    solver.solve();

    /* Generate files with constants */
    /*
    u2d_ScalarFieldOutput sourceFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/source", u2d_Constants::source(mesh->rr, gamma));
    u2d_ScalarFieldOutput avectorFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/avector", u2d_Constants::pythagoras(rr_dirichlet));
    u2d_ScalarFieldOutput bvectorFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/bvector", pyt_alpha);
    u2d_ScalarFieldOutput kvectorFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/kvector", u2d_Constants::thermalConductivity(mesh->rr));
    u2d_VectorFieldOutput bfieldFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/bfield", field, 2);
    u2d_VectorFieldOutput vfieldFileoutput("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square/vfield", u2d_Constants::advection(mesh->rr), 2);
    sourceFileoutput.output();
    avectorFileoutput.output();
    bvectorFileoutput.output();
    kvectorFileoutput.output();
    bfieldFileoutput.output();
    vfieldFileoutput.output();
    cout << " --== Source Vector ==--" << endl;
    cout << u2d_Constants::source(mesh->rr, gamma) << endl;
    cout << " --== a Vector ==--" << endl;
    cout << u2d_Constants::pythagoras(rr_dirichlet) << endl;
    cout << " --== b Vector ==--" << endl;
    cout << pyt_alpha << endl;
    cout << " --== k Vector ==--" << endl;
    cout << u2d_Constants::thermalConductivity(mesh->rr) << endl;
    cout << " --== b Field ==--" << endl;
    cout << field << endl;
    cout << " --== v Field ==--" << endl;
    cout << u2d_Constants::advection(mesh->rr) << endl;
    */



    /* Debug output */
    // cout << num_rows(rrb) << ", " << num_cols(rrb) << endl;
    // cout << problem.A << endl;
    // cout << problem.b << endl;
    //std::cout << problem.getSolution() << std::endl;
    //std::cout << u2d_Constants::pythagoras(mesh->rr) << std::endl; // solution in a square
    dense_vector<double> sol_error(num_rows(problem.getSolution()), 0.0);
    sol_error = problem.getSolution() - u2d_Constants::pythagoras(mesh->rr);
    std::cout << sol_error << std::endl;


    /* Output a plot file */
    u2d_PlotmtvOutput plotmtvout("solution", mesh->rr, mesh->jjd, problem.getSolution());
    u2d_PlotmtvOutput plotmtvout2("solution_exact", mesh->rr, mesh->jjd, u2d_Constants::pythagoras(mesh->rr));
    dense_vector<double> error(problem.getSize(), 0.0);
    error = u2d_Constants::pythagoras(mesh->rr) - problem.getSolution();
    u2d_PlotmtvOutput plotmtvout3("solution_error", mesh->rr, mesh->jjd, error);
    plotmtvout.output();
    plotmtvout2.output();
    plotmtvout3.output();
    delete mesh;
}


/**
 * Run a self contained test for a Helmholtz problem on a square with known
 * analytical solution. The Difference from exact solution is printed on screen.
 * This is the same as SquareTest just tailored with more debug output to
 * examine each term independently.
 * @author Gabriele Dini Ciacci <dark.schneider@iol.it>
 */
void u2d_Helmholtz::SquareTest_debug() {
    /* Load the mesh */
    u2d_Load_Umesh2d mesh_loader("/home/gab/projects/umesh2dgui/trunk/test/uhelm2d_square", "square_c", true);
    u2d_Mesh *mesh = mesh_loader.getMesh();
    /* Calculate the iso elements and the domain and boundary gauss points */
    u2d_IsoElement2DP13Points ie2d;
    u2d_IsoElement1DP12Points ie1d;
    u2d_DomainGaussPoints *dgp = ie2d.applyTo(mesh);
    u2d_BoundaryGaussPoints *bgp = ie1d.applyTo(mesh);
    /* Calculate Dirichlet Boundary Condition */
    u2d_DirichletCondition dc;
    dc.setSide(1);
    dc.setSide(2);
    dc.setSide(3);
    dc.generateNodes(mesh);
    /* Calculate Dirichlet Von Neumann Condition */
    u2d_NeumannCondition nc;
    nc.setSide(4);
    nc.calculateElements(mesh);


    //u2d_Constants::thermalConductivity(mesh->rr);  OK!!!
    //u2d_Constants::advection(mesh->rr); OK!!!
    //u2d_Constants::pythagoras(mesh->rr); OK!!!
    //u2d_Constants::gradThermalConductivity(mesh->rr); OK!!!
    //u2d_Constants::gradPythagoras(mesh->rr); OK!!!
    //u2d_Constants::source(mesh->rr, gamma); OK!!!
    //u2d_Constants::laplacianSolution(mesh->rr); OK!!!
    // field is OK!!!


    /* To check dirichlet */
    //    dense2D<double> rr_dirichlet2(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    //    std::set<int>::iterator it2;
    //    int rr_d_index2 = 0;
    //    for (it2 = dc.dirichlet_unique_nodes.begin();
    //            it2 != dc.dirichlet_unique_nodes.end();
    //            it2++) {
    //        printf("%d\n", *it2);
    //        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //            rr_dirichlet2(j, rr_d_index2) = mesh->rr(j, *it2 - 1);
    //        }
    //        ++rr_d_index2;
    //    }
    /* To check field */
    //    dense2D<double> rrb2(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    //    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
    //        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //            rrb2(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
    //        }
    //    }
    //    dense_vector<double> thermal_k2(num_cols(rrb2));
    //    thermal_k2 = u2d_Constants::thermalConductivity(rrb2);
    //    dense2D<double> spread_thermal_k2(2, num_rows(thermal_k2));
    //    spread_thermal_k2[0][mtl::iall] = trans(thermal_k2);
    //    spread_thermal_k2[1][mtl::iall] = trans(thermal_k2);
    //    dense2D<double> grad_pyt2(mesh->getSpatialDimension(), num_cols(rrb2));
    //    grad_pyt2 = u2d_Constants::gradPythagoras(rrb2);
    //    dense2D<double> field2(mesh->getSpatialDimension(), num_cols(rrb2));
    //    field2 = ele_prod(spread_thermal_k2, grad_pyt2);


    //    dense_vector<double> vect = u2d_Constants::laplacianSolution(mesh->rr);
    //    for (int i = 0; i < num_rows(vect); i++) {
    //        printf("%f\n", vect(i));
    //    }

    //    dense2D<double> mat = field2;
    //    for (int i = 0; i < num_rows(mat); i++) {
    //        for (int j = 0; j < num_cols(mat); j++) {
    //            printf("%f \t %f\n", mat(0, j), mat(1, j));
    //        }
    //    }

    /* To check RHS and LHS terms */
    u2d_LinearProblem<dense2D<double> > problem(mesh->getDomainNodeNumber());
    u2d_ProblemCalculator calculator(&problem, mesh);

    /* To check addDW__AlphaDG_left */
    calculator.addDW__AlphaDG_left(dgp, u2d_Constants::thermalConductivity(mesh->rr));

    /* To check addW__AlphaDG_left */
    calculator.addW__AlphaDG_left(dgp, u2d_Constants::advection(mesh->rr));

    /* To check addAlphaW__G_left */
    double gamma = 10.0;
    calculator.addAlphaW__G_left(dgp, gamma);

    /* To check addAlphaW__G_surf_left */
    double alpha = 2.0;
    calculator.addAlphaW__G_surf_left(bgp, &nc, alpha);

    /* To check forceDirichlet_left */
    calculator.forceDirichlet_left(&dc);

    /* To check addW__G_right */
    //double gamma = 10.0;
    calculator.addW__G_right(dgp, u2d_Constants::source(mesh->rr, gamma));

    /* To check addW__G_surf_right */
    //double alpha = 2.0;
    dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
        }
    }
    ///* Prepare the solution for addW__G_right on boundary */
    dense_vector<double> pyt_alpha(mesh->getBoundaryNodeNumber());
    pyt_alpha = alpha * u2d_Constants::pythagoras(rrb);
    calculator.addW__G_surf_right(bgp, &nc, pyt_alpha);

    /* To check addW__divG_surf_right */
    /* Prepare the field for addW__divG_right on the boundary */
    //dense2D<double> rrb(mesh->getSpatialDimension(), mesh->getBoundaryNodeNumber());
    //for (int i = 0; i < mesh->getBoundaryNodeNumber(); i++) {
    //    for (int j = 0; j < mesh->getSpatialDimension(); j++) {
    //        rrb(j, i) = mesh->rr(j, mesh->jjb_sd(i) - 1);
    //    }
    //}
    dense_vector<double> thermal_k(num_cols(rrb));
    thermal_k = u2d_Constants::thermalConductivity(rrb);
    dense2D<double> spread_thermal_k(2, num_rows(thermal_k));
    spread_thermal_k[0][mtl::irange()] = trans(thermal_k);
    spread_thermal_k[1][mtl::irange()] = trans(thermal_k);
    dense2D<double> grad_pyt(mesh->getSpatialDimension(), num_cols(rrb));
    grad_pyt = u2d_Constants::gradPythagoras(rrb);
    dense2D<double> field(mesh->getSpatialDimension(), num_cols(rrb));
    field = ele_prod(spread_thermal_k, grad_pyt);
    calculator.addW__divG_surf_right(bgp, &nc, field);

    /* To check forceDirichlet_right */
    /* Build a matrix with coordinates of boundary Dirichlet nodes */
    dense2D<double> rr_dirichlet(mesh->getSpatialDimension(), dc.dirichlet_unique_nodes.size());
    std::set<int>::iterator it;
    int rr_d_index = 0;
    for (it = dc.dirichlet_unique_nodes.begin();
            it != dc.dirichlet_unique_nodes.end();
            it++) {
        for (int j = 0; j < mesh->getSpatialDimension(); j++) {
            rr_dirichlet(j, rr_d_index) = mesh->rr(j, *it - 1);
        }
        ++rr_d_index;
    }
    calculator.forceDirichlet_right(&dc, u2d_Constants::pythagoras(rr_dirichlet));


#define vectorout 0
#define matrixout 1


#ifdef vectorout
    dense_vector<double> vect = problem.b;
    double total = 0.0;
    for (size_t i = 0; i < num_rows(vect); i++) {
        printf("%lu \t %e\n", i, vect(i));
        total += vect(i);
    }
    printf("%e\n", total);
#endif

#ifdef matrixout
    dense2D<double> mat = problem.A;
    double total_mat = 0.0;
    for (size_t i = 0; i < num_rows(mat); i++) {
        for (size_t j = 0; j < num_cols(mat); j++) {
            if (mat(i, j) != 0) {
                printf("%lu \t %lu \t %e\n", i, j, mat(i, j));
                total_mat += mat(i,j);
            }
        }
    }
    printf("%e\n", total_mat);
#endif
    //cout << problem.A << endl;

    delete mesh;
    return;

}
