/*   umesh2dgui, a GUI for umesh2d grid generator 

    Copyright (C) 2003 Gabriele Dini Ciacci
    Copyright (C) 2004 Gabriele Dini Ciacci
    Copyright (C) 2005 Gabriele Dini Ciacci
    Copyright (C) 2006 Gabriele Dini Ciacci
    Copyright (C) 2007 Gabriele Dini Ciacci
    Copyright (C) 2008 Gabriele Dini Ciacci
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/




#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
/* included for realpath() */
#include <limits.h>
/* for getcwd() */
#include <unistd.h>
#include <errno.h>
/* for stat() */
#include <sys/types.h>
#include <sys/stat.h>

/* Fltk */
#include <fltk/xpmImage.h>
#include <fltk/Browser.h>
#include <fltk/MultiBrowser.h>
#include <fltk/run.h>
#include <fltk/Button.h>
#include <fltk/Window.h>
#include <fltk/Group.h>
#include <fltk/CheckButton.h>
#include <fltk/Output.h>
#include <fltk/Color.h>
#include <fltk/filename.h>
#include <fltk/RadioButton.h>
#include <fltk/RadioItem.h>
#include <fltk/TextDisplay.h>

/* GUI and core part */
#include "umesh2dgui.h"
#include "u2d_core.h"
#include "u2d_GL.h"
#include "u2d_GLmesh.h"
#include "exceptions.h"
#include "common/u2d_Project.h"
#include "file_output/u2d_file_output.h"
#include "file_output/u2d_umesh2d_output.h"
#include "file_output/u2d_PlotmtvOutput.h"
#include "loaders/u2d_load_plotmtv.h"

/* Various */
#include "ascii_color.h"

/* Images */
#include "images/open_icon.xpm"
#include "images/save_7.xpm"
#include "images/cancel_15.xpm"
#include "images/create_12.xpm"

/* Solver */
#include "solver/uhelm2d/u2d_Helmholtz.h"
#include "solver/u2d_Problem.h"

/* tmp */
#include "u2d_load_scalarfield.h"
#include "u2d_load_vectorfield.h"
#include "boost/numeric/mtl/mtl.hpp"
#include "u2d_Mesh.h"
#include "u2d_load_umesh2d.h"
#include "solver/uhelm2d/u2d_Helmholtz.h"
#include "loaders/u2d_load_scalarfield.h"
#include "loaders/u2d_load_vectorfield.h"
#include "solver/u2d_Constants.h"
#include "common/u2d_Math.h"
#include "gl/u2d_DisplayText.h"

#define COLUMN1 50
#define COLUMN2 190
#define COLUMN3 310
#define COLUMN4 450

#define BROWSER_W 640
#define BROWSER_H 240
#define GLVIEW_W 202
#define GLVIEW_H 202
#define TITLEBAR_OFFSET 35
#define MANAGER_BORDER 5

#define MSGBUF_LENGTH 1024

/* Fixed for now, to move in configure */
#define USE_PNG 1

#ifdef USE_PNG 
#include <fltk/SharedImage.h>
#define LOAD_ICON_IMAGE(img) fltk::pngImage::get(img)
#else
/* Remember to include here the xpm image */
#define LOAD_ICON_IMAGE(img) fltk::xpmImage(img)
#endif


/* the file format requested, to be moved to class when redesigning */
int fileformat = FORMAT_UMESH2D;
/* a buffer for the message bar */
char msgbuf[MSGBUF_LENGTH];
u2d_Umesh2dgui *umesh2dgui;

int main(int argc, char ** argv) {
    int val = 0;
    /* PRINT SOME ABOUT INFO TO STDOUT */
    fprintf(stdout, "%s%s%s by %sGabriele Dini Ciacci%s - version %s%s%s\n",  \
  ASC_RED, PACKAGE, ASC_RESET, ASC_GREEN, ASC_RESET, ASC_BLUE, VERSION, ASC_RESET);

    /* PARSE THE ARGUMENTS WITH FLTK INTERNAL FUNCTION */
    fltk::args(argc, argv);

    //debug_cb();
    //exit(0);
    
    u2d_Umesh2dgui *umesh2dgui = new u2d_Umesh2dgui();

    umesh2dgui->showWindows(argc, argv);
    umesh2dgui->repositionWindows();
    val = umesh2dgui->loop();

    return val;

}

u2d_Umesh2dgui::u2d_Umesh2dgui() :
    ydataposition(0),
    ysteinerposition(0),
    ybackgridposition(0),
    project(NULL),
    mesh(NULL) {
    /* CLASS DECLARATION */
  u2d_umesh2d_output = new u2d_Umesh2dOutput;
  u2d_core = new u2d_Core;
  window = new fltk::Window(640, 534, "umesh2dgui");
  opengl_window = new fltk::Window(GLVIEW_W, GLVIEW_H, "OpenGL preview");
  prim_browser_window = new fltk::Window(BROWSER_W, BROWSER_H, "Primitive browser");

  allocateIcons();
  createMainWindow();
  createPreviewWindow();
  createBrowserWindow();
}

u2d_Umesh2dgui::~u2d_Umesh2dgui() {
    destroyMainWindow();
    destroyBrowserWindow();
    destroyPreviewWindow();
    destoryIcons();
    
    delete this->project;
    delete mesh;
    delete u2d_umesh2d_output;
    delete u2d_core;
}

void u2d_Umesh2dgui::allocateIcons() {
  /* ICON IMAGE DECLATATION */
  open_icon = new fltk::xpmImage(open_icon_xpm);
  save_icon = new fltk::xpmImage(save_7_xpm);
  create_icon = new fltk::xpmImage(create_12_xpm);
  cancel_icon = new fltk::xpmImage(cancel_15_xpm);
}

void u2d_Umesh2dgui::createMainWindow() {
  /* MAIN WINDOW SETUP */
  window->callback((fltk::Callback *)QuitItemCB, this);
  window->begin();
  menubar = new fltk::MenuBar(0, 0, 640, 30, 0);
  mainwindowtabgroup = new fltk::TabGroup(10, 30, 620, 480, 0);
  msgbar = new fltk::Output(0, 510, 640, 24);
  //  msgbar->deactivate();
  msgbar->color(fltk::GRAY75);
  outputMessage("Wellcome to the creation wizard! Select a primitive to start");
  window->end();

  createMainWindowTab();
  createMainWindowMenuBar();
}

void u2d_Umesh2dgui::createPreviewWindow() {
    /* OPENGL PREVIEW WINDOW SETUP*/
    opengl_window->begin();
    u2d_gl = new u2d_GL(1, 1, GLVIEW_W - 2, GLVIEW_H - 2, 0);
    opengl_window->end();
    /* Setup resizable */
    opengl_window->resizable(u2d_gl);
}

void u2d_Umesh2dgui::createBrowserWindow() {
    /* PRIMITIVE BROWSER WINDOW SETUP */
    const char *prim_browser_labels[] = {
        "Id",
        "Primitive Name",
        "Obj",
        "Type",
        "x1",
        "y1",
        "x2",
        "y2",
        "radius",
        "S arc",
        "E arc",
        0
    };
    int prim_browser_widths[] = {
        20,
        -1,
        20,
        50,
        70,
        70,
        70,
        70,
        70,
        50,
        50,
        0
    };
    prim_browser_window->begin();
    prim_browser = new fltk::MultiBrowser(0, 0, prim_browser_window->w(),    \
					prim_browser_window->h() - 45,    \
					"Primitive List");
    prim_browser->column_widths(prim_browser_widths);
    prim_browser->column_labels(prim_browser_labels);

    fltk::Button *browser_del_prim_button = new fltk::Button(10, prim_browser_window->h() - 35, 70, 25, "Delete");
    browser_del_prim_button->callback((fltk::Callback *)DeletePrimitiveFromBrowserButtonCB, this);
    prim_browser_window->end();
    /* Setup resizable */
    prim_browser_window->resizable(prim_browser_window);
}

void u2d_Umesh2dgui::createMainWindowTab() {
    /* TAB BAR SETUP */
    //  mainwindowtabgroup->box(fltk::UP_BOX);
    //  mainwindowtabgroup->box(fltk::BORDER_BOX);
    mainwindowtabgroup->box(fltk::THIN_UP_BOX);
    mainwindowtabgroup->begin(); // 620 450
    boundarygroup = new fltk::Group(0, 30, 620, 440, "Boundary");
    boundarygroup->box(fltk::ENGRAVED_BOX);
    domaingroup = new fltk::Group(0, 30, 620, 440, "Domain");
    domaingroup->box(fltk::EMBOSSED_BOX);
    creationgroup = new fltk::Group(0, 30, 620, 440, "Creation");
    creationgroup->box(fltk::EMBOSSED_BOX);
    outputgroup = new fltk::Group(0, 30, 620, 440, "Output");
    outputgroup->box(fltk::EMBOSSED_BOX);
    meshgroup = new fltk::Group(0, 30, 620, 440, "Mesh");
    meshgroup->box(fltk::EMBOSSED_BOX);
    sidesgroup = new fltk::Group(0, 30, 620, 440, "Sides");
    sidesgroup->box(fltk::EMBOSSED_BOX);
    parametersgroup = new fltk::Group(0, 30, 620, 440, "Parameters");
    parametersgroup->box(fltk::EMBOSSED_BOX);
    mainwindowtabgroup->end();
    /* Create the tabs */
    createMainWindowBoundaryTab();
    createMainWindowDomainTab();
    createMainWindowCreationTab();
    createMainWindowOutputTab();
    createMainWindowMeshTab();
    createMainWindowSidesTab();
    createMainWindowParametersTab();
}


void u2d_Umesh2dgui::createMainWindowBoundaryTab() {
      /* BOUNDARY TAB */
  boundarygroup->begin();

  /* next (and prev) button */
  fltk::Button *boundary_nextbutton = new fltk::Button(320, 400, 100, 25, "Next  ");
  boundary_nextbutton->callback((fltk::Callback *)BoundaryTabNextButtonCB, this);
  boundary_nextbutton->image(LOAD_ICON_IMAGE("images/next.png"));
  boundary_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  boundary_nextbutton->tooltip("Proceed to domain definition");

  /* the subwindow */
  //  boundary_subwindow = new fltk::Group(10, 50, 600, 340, "Insert primitive data here:");
  boundary_subwindow = new fltk::Group(10, 50, 600, 340, "");
  boundary_subwindow->box(fltk::SHADOW_BOX);

  /* PRIMITIVE MENU SETUP */
  /*  this is inside boundary tab */

  /* Choice */
  /*
   *  fltk::Choice *primitivemenu = new fltk::Choice(210, 10, 100, 20,\
   *    "Choose type of primitive to add:");
   *  primitivemenu->add("&clear", 0, (fltk::Callback *)clear_cb, 0, 0);
   *  primitivemenu->add("&line", 0, (fltk::Callback *)line_cb, 0, 0);
   *  primitivemenu->add("&circle", 0, (fltk::Callback *)circle_cb, 0, 0);
   *  primitivemenu->add("&ellipse", 0, (fltk::Callback *)ellipse_cb, 0, 0);
   *  primitivemenu->add("&data", 0, (fltk::Callback *)clear_cb, 0, 0);
   */

  fltk::MenuBar *primitivemenu = new fltk::MenuBar(5, 2, 600, 30);

  primitivemenu->begin();
  lineitem = new fltk::Item("Line");
  lineitem->callback((fltk::Callback*)ShowLineCB, this);
  lineitem->image(LOAD_ICON_IMAGE("images/line_1.png"));
  circleitem = new fltk::Item("Circle");
  circleitem->callback((fltk::Callback*)ShowCircleCB, this);
  circleitem->image(LOAD_ICON_IMAGE("images/circle_1.png"));
  ellipseitem = new fltk::Item("Ellipse");
  ellipseitem->callback((fltk::Callback*)ShowEllipseCB, this);
  ellipseitem->image(LOAD_ICON_IMAGE("images/ellipse_1.png"));
  dataitem = new fltk::Item("Data");
  dataitem->callback((fltk::Callback*)ShowDataCB, this);
  dataitem->image(LOAD_ICON_IMAGE("images/data_1.png"));
  rectangleitem = new fltk::Item("Rectangle");
  rectangleitem->callback((fltk::Callback*)ShowRectangleCB, this);
  rectangleitem->image(LOAD_ICON_IMAGE("images/rectangle_2.png"));
  primitivemenu->end();

  boundarygroup->end();

}

void u2d_Umesh2dgui::createMainWindowDomainTab() {
      /* DOMAIN TAB */
  domaingroup->begin();

  /* next and prev buttons */
  fltk::Button *domain_nextbutton = new fltk::Button(320, 400, 100, 25, "Next  ");
  domain_nextbutton->callback((fltk::Callback *)DomainTabNextButtonCB, this);
  domain_nextbutton->image(LOAD_ICON_IMAGE("images/next.png"));
  domain_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  domain_nextbutton->tooltip("Proceed to file output");

  fltk::Button *domain_prevbutton = new fltk::Button(200, 400, 100, 25, "  Prev");
  domain_prevbutton->callback((fltk::Callback *)DomainTabPrevButtonCB, this);
  domain_prevbutton->image(LOAD_ICON_IMAGE("images/prev.png"));
  domain_prevbutton->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  domain_prevbutton->tooltip("Back to boundary definition");

  domain_subwindow = new fltk::Group(10, 50, 600, 340, "");
  domain_subwindow->box(fltk::SHADOW_BOX);

  /* the subwindow */
  domain_subwindow->begin();

  /* backgrid bordered box*/
  fltk::Group *backgridcontour  = new fltk::Group(10, 10, 580, 145, "");
  backgridcontour->box(fltk::BORDER_FRAME);
  backgridcontour->align(fltk::ALIGN_INSIDE_LEFT);

  backgridcontour->begin();

  fltk::Button *new_backgrid = new fltk::Button(450, 115, 120, 20, "Add backgrid point");
  new_backgrid->callback((fltk::Callback *)AddBackgridLineCB, this);

  backgridgroup = new fltk::ScrollGroup(10, 10, 560, 100, "");
  backgridgroup->scrollbar.linesize(20);

  backgridcontour->end();

  /* steiner bordered box*/
  fltk::Group *steinercontour  = new fltk::Group(10, 170, 580, 145, "");
  steinercontour->box(fltk::BORDER_FRAME);
  steinercontour->align(fltk::ALIGN_INSIDE_LEFT);

  steinercontour->begin();

  fltk::Button *new_steiner = new fltk::Button(450, 115, 120, 20, "Add steiner point");
  new_steiner->callback((fltk::Callback *)AddSteinerLineCB, this);

  steinergroup = new fltk::ScrollGroup(10, 10, 560, 100, "");
  steinergroup->scrollbar.linesize(20);

  steinercontour->end();

  domain_subwindow->end();

  domaingroup->end();
}

void u2d_Umesh2dgui::createMainWindowCreationTab() {
  /* CREATION TAB */
  creationgroup->begin();

  /* next and prev buttons */
  fltk::Button *creation_prevbutton = new fltk::Button(200, 400, 100, 25, "  Prev");
  creation_prevbutton->callback((fltk::Callback *)CreationTabPrevButtonCB, this);
  creation_prevbutton->image(LOAD_ICON_IMAGE("images/prev.png"));
  creation_prevbutton->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
  creation_prevbutton->tooltip("Back to domain definition");

  fltk::Button *creation_nextbutton = new fltk::Button(320, 400, 100, 25, "Finish  ");
  creation_nextbutton->callback((fltk::Callback *)CreationNextButtonCB, this);
  creation_nextbutton->image((fltk::Image *)create_icon);
  creation_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
  creation_nextbutton->tooltip("Output file and generate mesh");

  creation_subwindow = new fltk::Group(10, 50, 600, 340, "");
  creation_subwindow->box(fltk::SHADOW_BOX);

  creation_subwindow->begin();

  /* output format bordered box */
  fltk::Group *outcontour  = new fltk::Group(COLUMN1, 31, 240, 67, "Select output\n     format");
  outcontour->box(fltk::BORDER_FRAME);
  outcontour->align(fltk::ALIGN_INSIDE_LEFT);

  outcontour->begin();
  fltk::RadioButton *umesh2d_format = new fltk::RadioButton(130, 3, 98, 20, "umesh2d");
  umesh2d_format->callback((fltk::Callback *)SetUmesh2dFormatCB, this);
  umesh2d_format->value(true);
  fltk::RadioButton *umesh2d_fe_format = new fltk::RadioButton(130, 21, 98, 20, "umesh2d FE");
  umesh2d_fe_format->callback((fltk::Callback *)SetUmesh2dFEFormatCB, this);
  fltk::RadioButton *umesh2d_fv_format = new fltk::RadioButton(130, 39, 98, 20, "umesh2d FV");
  umesh2d_fv_format->callback((fltk::Callback *)SetUmesh2dFVFormatCB, this);
  outcontour->end();


  /* Filename */
  fltk::Input *filenameinput = new fltk::Input(110, 300, 250, 25, "Project name :");
  //  filenameinput->box(fltk::SHADOW_BOX);
  filenameinput_i = creation_subwindow->find(filenameinput);

  /* Filename browser button */
  fltk::Button *filenamebrowser = new fltk::Button(362, 300, 25, 25, "");
  filenamebrowser->callback((fltk::Callback *)OpenFileBrowserCB, this);
  filenamebrowser->image((fltk::Image *)open_icon);
  filenamebrowser->tooltip("Browse");
  //  filenamebrowser->buttonbox(fltk::SHADOW_BOX);

  creation_subwindow->end();

  creationgroup->end();
}

void u2d_Umesh2dgui::createMainWindowOutputTab() {
    /* OUTPUT TAB */
  outputgroup->begin();
  shell_output = new fltk::TextDisplay(1, 1, 620, 440);
  outputgroup->end();
}

void u2d_Umesh2dgui::createMainWindowMeshTab() {
      /* MESH TAB */
  meshgroup->begin();

  /* opengl mesh window setup */
  u2d_glmesh = new u2d_GLmesh(1, 1, 620, 440, 0);
  meshgroup->end();
}

void u2d_Umesh2dgui::createMainWindowSidesTab() {
    sidesgroup->begin();

    fltk::Button *sides_nextbutton = new fltk::Button(320, 400, 100, 25, "Next  ");
    sides_nextbutton->callback((fltk::Callback *)SidesTabNextButtonCB, this);
    sides_nextbutton->image(LOAD_ICON_IMAGE("images/next.png"));
    sides_nextbutton->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
    sides_nextbutton->tooltip("Proceed to set problem parameters");

    sides_subwindow = new fltk::Group(10, 50, 600, 340, "");
    sides_subwindow->box(fltk::SHADOW_BOX);

    sides_subwindow->begin();

    /* Mesh info bordered box */
    outcontour_meshinfo = new fltk::Group(COLUMN1, 9, 500, 59, "Mesh\ninfo");
    outcontour_meshinfo->box(fltk::BORDER_FRAME);
    outcontour_meshinfo->align(fltk::ALIGN_INSIDE_LEFT);

    /*
     * sides
     * boundary nodes n
     * domain nodes n
     * domain elements n
     * boundary elements n
     */
    outcontour_meshinfo->begin();
    domainnodesoutput = new fltk::Output(190, 3, 50, 25, "D. Nodes :");
    boundarynodesoutput = new fltk::Output(190, 32, 50, 25, "B. Nodes :");

    domainelementsoutput = new fltk::Output(310, 3, 50, 25, "D. Elem. :");
    boundaryelementsoutput = new fltk::Output(310, 32, 50, 25, "B. Elem. :");

    sidesoutput = new fltk::Output(420, 3, 50, 25, "Sides :");
    //fltk::Output *<yalla>soutput = new fltk::Output(300, 32, 50, 25, "B. Elem. :");
    outcontour_meshinfo->end();

    /* Sides info bordered box */
    outcontour_sidesinfo = new fltk::Group(COLUMN1, 70, 500, 230, "Sides\ninfo");
    outcontour_sidesinfo->box(fltk::BORDER_FRAME);
    outcontour_sidesinfo->align(fltk::ALIGN_INSIDE_LEFT);

    outcontour_sidesinfo->begin();

    outcontour_sidesinfo->end();

     /* Buttons */
    fltk::Button *load = new fltk::Button(200, 305, 100, 25, "Load mesh");
    //load->image((fltk::Image *)cancel_icon);
    load->tooltip("Load the mesh info");
    load->callback((fltk::Callback *)LoadMeshCB, this);

    sides_subwindow->end();

    sidesgroup->end();
}

void u2d_Umesh2dgui::createMainWindowParametersTab() {
    parametersgroup->begin();

    /* Buttons */
    fltk::Button *parameters_prevbutton = new fltk::Button(200, 400, 100, 25, "  Prev");
    parameters_prevbutton->callback((fltk::Callback *)ParametersTabPrevButtonCB, this);
    parameters_prevbutton->image(LOAD_ICON_IMAGE("images/prev.png"));
    parameters_prevbutton->align(fltk::ALIGN_LEFT | fltk::ALIGN_INSIDE);
    parameters_prevbutton->tooltip("Back to boundary condition type definition");

    fltk::Button *solve = new fltk::Button(320, 400, 100, 25, "Solve");
    //add->image(LOAD_ICON_IMAGE("images/solve.png"));
    solve->tooltip("Solve the problem");
    solve->align(fltk::ALIGN_RIGHT | fltk::ALIGN_INSIDE);
    solve->callback((fltk::Callback *)SolveProblemCB, this);

    conditions_subwindow = new fltk::Group(10, 50, 600, 340, "");
    conditions_subwindow->box(fltk::SHADOW_BOX);

    conditions_subwindow->begin();

    /* The problem formula */
    fltk::Widget *helmholtz_image_box = new fltk::Widget(220, 3, 190, 66, "Helmholtz equation: ");
    helmholtz_image_box->image(LOAD_ICON_IMAGE("images/helmholtz_problem_0.7.png"));
    /* If we want to use a transparent png */
    //helmholtz_image_box->image(LOAD_ICON_IMAGE("images/helmholtz_problem_70dpi_trasp.png"));
    //helmholtz_image_box->box(fltk::NO_BOX);
    helmholtz_image_box->align(fltk::ALIGN_LEFT);

    /* boundary conditions bordered box */
    fltk::Group *outcontour_boundarycondtions = new fltk::Group(COLUMN1, 70, 500, 114, "Boundary\nconditions");
    outcontour_boundarycondtions->box(fltk::BORDER_FRAME);
    outcontour_boundarycondtions->align(fltk::ALIGN_INSIDE_LEFT);

    outcontour_boundarycondtions->begin();
    avectorinput = new fltk::Input(190, 3, 190, 25, "a(s) vector :");
    bvectorinput = new fltk::Input(190, 30, 190, 25, "b(s) vector :");
    bfieldinput = new fltk::Input(190, 57, 190, 25, "b(s) field :");
    alphainput = new fltk::Input(190, 84, 190, 25, "alpha scalar :");

    /* Radio buttons to select formula or file*/
    for (int idx = 0; idx < 3; idx++) {
        fltk::Group *radiogroup = new fltk::Group(390, 5 + 27 * idx, 100, 20, "");
        /* Create the radiobuttons */
        radiogroup->begin();
        fltk::RadioButton *isequationradio = new fltk::RadioButton(0, 0, 50, 20, "Eqn");
        isequationradio->value(true); // Set equation as default
        fltk::RadioButton *isfileradio = new fltk::RadioButton(50, 0, 60, 20, "File");
        //isfileradio->value(true); // Set file as default
        radiogroup->end();
        this->vectorradiogroup.push_back(radiogroup);
    }        
    outcontour_boundarycondtions->end();

    /* parameters bordered box */
    fltk::Group *outcontour_problemparameters = new fltk::Group(COLUMN1, 186, 500, 144, "Problem\nparameters");
    outcontour_problemparameters->box(fltk::BORDER_FRAME);
    outcontour_problemparameters->align(fltk::ALIGN_INSIDE_LEFT);

    outcontour_problemparameters->begin();
    sourcevectorinput = new fltk::Input(190, 3, 190, 25, "source(r) vector :");
    kvectorinput = new fltk::Input(190, 30, 190, 25, "k(r) vector :");
    vxfieldinput = new fltk::Input(190, 57, 190, 25, "vx(r) field :");
    vyfieldinput = new fltk::Input(190, 84, 190, 25, "vy(r) field :");
    gammainput = new fltk::Input(190, 111, 190, 25, "gamma scalar :");
    
    /* Radio buttons to select formula or file*/
    for (int idx = 0; idx < 3; idx++) {
        fltk::Group *radiogroup = new fltk::Group(390, 5 + 27 * idx, 100, 20, "");
        /* Create the radiobuttons */
        radiogroup->begin();
        fltk::RadioButton *isequationradio = new fltk::RadioButton(0, 0, 50, 20, "Eqn");
        isequationradio->value(true); // Set equation as default
        fltk::RadioButton *isfileradio = new fltk::RadioButton(50, 0, 60, 20, "File");
        //isfileradio->value(true); // Set file as default
        radiogroup->end();
        this->vectorradiogroup.push_back(radiogroup);
    }  
    outcontour_problemparameters->end();

    conditions_subwindow->end();

    parametersgroup->end();
}

void u2d_Umesh2dgui::createMainWindowMenuBar() {
  /* MENUBAR SETUP */
  menubar->begin();
  /* file menu */
  fltk::ItemGroup *filemenu = new fltk::ItemGroup("&File");
  filemenu->begin();
  fltk::Item *openitem = new fltk::Item("&Open");
  openitem->callback((fltk::Callback *)OpenItemCB, this);
  fltk::Item *saveitem = new fltk::Item("&Save");
  saveitem->callback((fltk::Callback *)SaveItemCB, this);
  saveitem->image((fltk::Image *)save_icon);
  //  cancel->tooltip("Cancel");
  fltk::Item *saveasitem = new fltk::Item("&Save As...");
  saveasitem->callback((fltk::Callback *)SaveAsItemCB, this);
  fltk::Item *quititem = new fltk::Item("&Quit");
  quititem->callback((fltk::Callback *)QuitItemCB, this);
  filemenu->end();
  /* view menu */
  fltk::ItemGroup *view = new fltk::ItemGroup("&View");
  view->begin();
  fltk::Item *glview = new fltk::Item("&Toggle openGL preview");
  glview->callback((fltk::Callback *)GlViewItemCB, this);
  glview->type(fltk::Button::TOGGLE);
  fltk::Item *browserview = new fltk::Item("&Toggle Primirive browser");
  browserview->callback((fltk::Callback *)BrowserViewItemCB, this);
  browserview->type(fltk::Button::TOGGLE);
  view->end();
  /* tools menu */
  fltk::ItemGroup *toolsmenu = new fltk::ItemGroup("&Tools");
  toolsmenu->begin();
  fltk::Item *main_coreitem = new fltk::Item("&Evaluate closed objects");
  main_coreitem->callback((fltk::Callback *)RunMainCoreCB, this);
  toolsmenu->end();
  /* help menu */
  fltk::ItemGroup *help = new fltk::ItemGroup("&Help");
  help->begin();
  fltk::Item *run_helmholtz = new fltk::Item("Run &Helmholtz test");
  run_helmholtz->callback((fltk::Callback *)RunHelmholtzSquareTestCB, this);
  fltk::Item *about = new fltk::Item("&About umesh2dgui");
  about->callback((fltk::Callback *)AboutItemCB, this);

#ifdef DEBUGGING
  fltk::Item *debugitem = new fltk::Item("&Debug callback");
  debugitem->callback((fltk::Callback *)debug_cb, this);
  fltk::Item *debugguiitem = new fltk::Item("&Debug GUI callback");
  debugguiitem->callback((fltk::Callback *)DebugCB, this);
#endif

  help->end();
  menubar->end();
}

void u2d_Umesh2dgui::createMainWindowTitleBox() {
  /* TITLE BOX SETUP */ /* removed now */
  /*  fltk::Widget *box = new fltk::Widget(10, 30, 620, 40, "umesh2dgui");
   *  box->box(fltk::UP_BOX);
   *  box->labelfont(fltk::HELVETICA_BOLD_ITALIC);
   *  box->labelcolor(fltk::GREEN);
   *  box->labelsize(40);
   *  box->labeltype(fltk::SHADOW_LABEL);
   */
}

void u2d_Umesh2dgui::showWindows(int argc, char ** argv) {

    /* Show the main window */
    window->show(argc, argv);
    /* Select the main window boundary definition tab */
    boundary_subwindow->show();
    /* Select the line calback */
    showLineCB(boundary_subwindow, this); //NOT SURE IF OK TO PASS boundary_subwindow
    /* Show the opengl window */
    opengl_window->show();
    /* Show the primitive browser window */
    prim_browser_window->show();
}

void u2d_Umesh2dgui::repositionWindows() {
    int val;
    val = fltk::wait();
    /* Now that we got from system the window position we
     * reposition our windows in a nice way */
    prim_browser_window->position(window->x(), window->y() + window->h() + TITLEBAR_OFFSET);
    //  prim_browser_window->position(window->x(), window->y() - BROWSER_H - TITLEBAR_OFFSET);
    opengl_window->position(window->x() + window->w() + MANAGER_BORDER, window->y());
}

int u2d_Umesh2dgui::loop() {
    return fltk::run();
}

void u2d_Umesh2dgui::destroyMainWindow() {
  window->clear();
    delete window;
}

void u2d_Umesh2dgui::destroyBrowserWindow() {
  opengl_window->clear();
    delete opengl_window;
}

void u2d_Umesh2dgui::destroyPreviewWindow() {
  prim_browser_window->clear();
    delete prim_browser_window;
}

void u2d_Umesh2dgui::destoryIcons() {
  delete open_icon;
  delete save_icon;
  delete create_icon;
  delete cancel_icon;
}


void u2d_Umesh2dgui::openFileBrowserCB(fltk::Widget *widget, void *ptr) {
   ((fltk::Input *)creation_subwindow->child(filenameinput_i))->value(fltk::file_chooser("Select a file", NULL, NULL, true));
  /* try to call file browser several time. Value() wants a good static buffer or what?
   * LOOK LIKE WORKING... but i'm really not sure if will under extreme condition
   */
}

void u2d_Umesh2dgui::OpenFileBrowserCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->openFileBrowserCB(widget, ptr);
}

void u2d_Umesh2dgui::creationNextButtonCB(fltk::Widget *widget, void *ptr) {
    int returnvalue;
    if (strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1) {
        printf("Filename is empty. Opening file browser\n");
        openFileBrowserCB(widget, ptr);
    }
    if (strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1) {
        printf("Filename is empty. User aborted\n");
        return;
    }

    switch (fileformat) {
        case FORMAT_UMESH2D:
            u2d_umesh2d_output->setFormat(FORMAT_UMESH2D);
            break;
        case FORMAT_UMESH2D_FE:
            u2d_umesh2d_output->setFormat(FORMAT_UMESH2D_FE);
            break;
        case FORMAT_UMESH2D_FV:
            u2d_umesh2d_output->setFormat(FORMAT_UMESH2D_FV);
            break;
        default:
        {
            fprintf(stderr, "File format not supported\n");
            outputMessage("File format not supported");
            break;
        }
    }

    /* Copy steiner and backgrid info to u2d_core*/
    u2d_core->allocate_steiner(steinergroup->children() / 3);
#ifdef DEBUGGING
    printf("Steiners in widgets: %i\n", steinergroup->children() / 3);
#endif
    for (int i = 0; i < steinergroup->children(); i += 3) {
        u2d_core->set_steiner(i / 3, ((fltk::ValueInput *)steinergroup->child(i))->value(),
                ((fltk::ValueInput *)steinergroup->child(i + 1))->value());
    }

    u2d_core->allocate_backgrid(backgridgroup->children() / 4);
#ifdef DEBUGGING
    printf("Backgrid in widgets: %i\n", backgridgroup->children() / 4);
#endif
    for (int i = 0; i < backgridgroup->children(); i += 4) {
        u2d_core->set_backgrid(i / 4, ((fltk::ValueInput *)backgridgroup->child(i))->value(),
                ((fltk::ValueInput *)backgridgroup->child(i + 1))->value(),
                ((fltk::ValueInput *)backgridgroup->child(i + 2))->value());
    }

    /* We can now create a project */
    std::string projectNameInput(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value());
    try {
      this->setProject(u2d_Project::CreateProject(projectNameInput));
    }
    catch (u2d_IO_Exception &e) {
        fltk::alert(e.what());
        outputMessage("Project creation failed. Please try again.");
        return;
    }
       
    /* Set filenames in u2d_umesh2d_output */
    u2d_umesh2d_output->setFilename(project->getProjectAbsoluteDir(true));

    /* Calculate closed objects */
    fprintf(stdout, "Calculating closed objects and vertex enumeration...\n");
    outputMessage("Calculating closed objects and vertex enumeration...");
    u2d_core->find_objects();
    fprintf(stdout, "Done!\n");

    /* Initialize the u2d_file_output class with necessary parameters */
    u2d_umesh2d_output->select_prim_list(u2d_core->get_prim_list_start(), u2d_core->get_prim_list_end(),  \
			   u2d_core->get_prim_number(), u2d_core->get_vertex_n());
    u2d_umesh2d_output->setSteiners(u2d_core->steiners, u2d_core->steiner_n);
    u2d_umesh2d_output->setBackgrids(u2d_core->backgrids, u2d_core->backgrid_n);

    /* Call the output boundary file function of the
     * selected class for output */
    try {
        u2d_umesh2d_output->output();
    } catch (exception &e) {
        fltk::alert(e.what());
        outputMessage("Umesh2d output failed. Umesh2d output files have not been created.");
        return;
    }
    fprintf(stdout, "File creation successful\n");
    outputMessage("File creation successful");

    /* Show the output tab */
    creationgroup->hide();
    outputgroup->show();

    /* Run the mesher */
    executeMesher(project);
    drawMesh(project);

    /* Show the mesh tab */
    outputgroup->hide();
    meshgroup->show();
}

void u2d_Umesh2dgui::CreationNextButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->creationNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::showLineCB(fltk::Widget *widget, void *ptr) {
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  resetBoundaryTabCB(widget, ptr);
  lineitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* spline*/
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  //  splinepoints->pagesize(10.0);
  inputindex[0] = boundary_subwindow->find(splinepoints);
  //  printf("%i", inputindex[0]);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* start point bordered box */
  fltk::Group *scontour  = new fltk::Group(COLUMN1, 31, 240, 47, "Start point");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x1 */
  fltk::ValueInput *x1 = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x: ");
  x1->range(-1000, 1000);
  x1->step(0.01);
  x1->linesize(0.01);
  inputindex[2] = boundary_subwindow->find(x1);

  /* y1 */
  fltk::ValueInput *y1 = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y: ");
  y1->range(-1000, 1000);
  y1->step(0.01);
  y1->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(y1);
  //  printf("%i", inputindex[2]);


  /* end point bordered box */
  fltk::Group *econtour  = new fltk::Group(COLUMN3, 31, 240, 47, "End point");
  econtour->box(fltk::BORDER_FRAME);
  econtour->align(fltk::ALIGN_INSIDE_LEFT);

  /* x2 */
  fltk::ValueInput *x2 = new fltk::ValueInput(COLUMN4, 34, 98, 20, "x: ");
  x2->range(-1000, 1000);
  x2->step(0.01);
  x2->linesize(0.01);
  //  x2->pagesize(1.0);
  inputindex[4] = boundary_subwindow->find(x2);
  //  printf("%i", inputindex[3]);

  /* y2 */
  fltk::ValueInput *y2 = new fltk::ValueInput(COLUMN4, 55, 98, 20, "y: ");
  y2->range(-1000, 1000);
  y2->step(0.01);
  y2->linesize(0.01);
  //  y2->pagesize(1.0);
  inputindex[5] = boundary_subwindow->find(y2);

  /* boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 70, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[6] = boundary_subwindow->find(isboundary);
  /* Datagroup for [i|s|h]data */
  /*  fltk::Group *databox = new fltk::Group(10, 130, 580, 160, "");
  databox->box(fltk::BORDER_FRAME);
  */
  fltk::Button *new_dataline = new fltk::Button(10, 100, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)NewDataLineCB, this);
  //  datagroup = new fltk::ScrollGroup(10, 140, 580, 160, "Insert [i|s|h]data here:");
  datagroup = new fltk::ScrollGroup(10, 130, 580, 160, "");
  datagroup->scrollbar.linesize(20);
  newDataLineCB(widget, ptr);
  newDataLineCB(widget, ptr);
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  //  datagroup->box(fltk::BORDER_FRAME);
  //  datagroup->begin();
  //  fltk::Widget *box2 = new fltk::Widget(0, 0, 620, 160, "TEXT"); // this is to see the space avaiable
  //  fltk::Scroll *datascroll = new fltk::Scroll(600, 0, 20, 160, 0);
  //  datagroup->end();

  /* Buttons (Add, Cancel)*/
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)ResetBoundaryTabCB, this);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "&Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  add->tooltip("Add the primitive");
  add->callback((fltk::Callback *)AddLineCB, this);
  boundary_subwindow->end();
  boundary_subwindow->redraw();
}

void u2d_Umesh2dgui::ShowLineCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->showLineCB(widget, ptr);
}

void u2d_Umesh2dgui::showCircleCB(fltk::Widget *widget, void *ptr){
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  resetBoundaryTabCB(widget, ptr);
  circleitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();
  /* spline */
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  inputindex[0] = boundary_subwindow->find(splinepoints);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* radius */
  fltk::ValueInput *radius = new fltk::ValueInput(COLUMN2, 31, 100, 20, "Radius: ");
  radius->range(-1000, 1000);
  radius->step(0.01);
  inputindex[2] = boundary_subwindow->find(radius);

  /* center point bordered box */
  fltk::Group *ccontour  = new fltk::Group(COLUMN1, 52, 240, 47, "Center");
  ccontour->box(fltk::BORDER_FRAME);
  ccontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* center x */
  fltk::ValueInput *centerx = new fltk::ValueInput(COLUMN2, 55, 98, 20, "abscissa (x): ");
  centerx->range(-1000, 1000);
  centerx->step(0.01);
  centerx->linesize(0.01);
  inputindex[3] = boundary_subwindow->find(centerx);

  /* center y */
  fltk::ValueInput *centery = new fltk::ValueInput(COLUMN2, 76, 98, 20, "ordinata (y): ");
  centery->range(-1000, 1000);
  centery->step(0.01);
  centery->linesize(0.01);
  inputindex[4] = boundary_subwindow->find(centery);


  /* end point bordered box */
  fltk::Group *acontour  = new fltk::Group(COLUMN3, 52, 240, 47, "Angle (deg)");
  acontour->box(fltk::BORDER_FRAME);
  acontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* start arc */
  fltk::ValueInput *sarc = new fltk::ValueInput(COLUMN4, 55, 98, 20, "start:");
  sarc->range(-360, 360);
  sarc->linesize(1.0);
  sarc->step(0.1);
  inputindex[5] = boundary_subwindow->find(sarc);

  /* end arc */
  fltk::ValueInput *earc = new fltk::ValueInput(COLUMN4, 76, 98, 20, "end:");
  earc->range(-360, 360);
  earc->linesize(1.0);
  earc->step(0.1);
  inputindex[6] = boundary_subwindow->find(earc);

  /* is boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 90, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[7] = boundary_subwindow->find(isboundary);
  /* [i|s|h]data */
  fltk::Button *new_dataline = new fltk::Button(10, 120, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)NewDataLineCB, this);
  datagroup = new fltk::ScrollGroup(10, 150, 580, 140, "");
  datagroup->scrollbar.linesize(20);
  newDataLineCB(widget, ptr);
  newDataLineCB(widget, ptr);
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  /* buttons */
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)ResetBoundaryTabCB, this);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)AddCircleCB, this);
  boundary_subwindow->end();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 380);
}

void u2d_Umesh2dgui::ShowCircleCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->showCircleCB(widget, ptr);
}

void u2d_Umesh2dgui::showEllipseCB(fltk::Widget *widget, void *ptr) {
  /* POPULATE THE BOUNDARY_SUBWINDOW */
  resetBoundaryTabCB(widget, ptr);
  ellipseitem->box(fltk::DOWN_BOX);
  boundary_subwindow->begin();

  /* spline */
  fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
  splinepoints->range(0, 10000);
  splinepoints->linesize(1.0);
  splinepoints->step(1);
  splinepoints->set_value(100);
  inputindex[0] = boundary_subwindow->find(splinepoints);

  /* name */
  fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
  inputindex[1] = boundary_subwindow->find(primname);

  /* Semixais Length */
  fltk::Group *scontour  = new fltk::Group(COLUMN1, 31, 240, 47, "Semiaxis length");
  scontour->box(fltk::BORDER_FRAME);
  scontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* semix */
  fltk::ValueInput *semix = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x:");
  semix->range(-1000, 1000);
  semix->step(0.01);
  inputindex[2] = boundary_subwindow->find(semix);

  /* semiy */
  fltk::ValueInput *semiy = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y:");
  semiy->range(-1000, 1000);
  semiy->step(0.01);
  inputindex[3] = boundary_subwindow->find(semiy);

  /* center point bordered box */
  fltk::Group *ccontour  = new fltk::Group(COLUMN1, 80, 240, 47, "Center");
  ccontour->box(fltk::BORDER_FRAME);
  ccontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* center x */
  fltk::ValueInput *centerx = new fltk::ValueInput(COLUMN2, 83, 98, 20, "abscissa (x): ");
  centerx->range(-1000, 1000);
  centerx->step(0.01);
  centerx->linesize(0.01);
  inputindex[4] = boundary_subwindow->find(centerx);

  /* center y */
  fltk::ValueInput *centery = new fltk::ValueInput(COLUMN2, 104, 98, 20, "ordinata (y): ");
  centery->range(-1000, 1000);
  centery->step(0.01);
  centery->linesize(0.01);
  inputindex[5] = boundary_subwindow->find(centery);


  /* Angle */
  fltk::Group *acontour  = new fltk::Group(COLUMN3, 31, 240, 47, "Angle (deg)");
  acontour->box(fltk::BORDER_FRAME);
  acontour->align(fltk::ALIGN_INSIDE_LEFT);

  /* start arc */
  fltk::ValueInput *sarc = new fltk::ValueInput(COLUMN4, 34, 98, 20, "start:");
  sarc->range(-360, 360);
  sarc->linesize(1.0);
  sarc->step(0.1);
  inputindex[6] = boundary_subwindow->find(sarc);

  /* end arc */
  fltk::ValueInput *earc = new fltk::ValueInput(COLUMN4, 55, 98, 20, "end:");
  earc->range(-360, 360);
  earc->linesize(1.0);
  earc->step(0.1);
  inputindex[7] = boundary_subwindow->find(earc);

  /* is boundary */
  //  fltk::CheckButton *isboundary = new fltk::CheckButton(200, 90, 100, 20, "Boundary: ");
  //  isboundary->align(fltk::ALIGN_LEFT);
  //  inputindex[8] = boundary_subwindow->find(isboundary);
  /* [i|s|h]data */
  fltk::Button *new_dataline = new fltk::Button(10, 140, 580, 20, "Add data entry");
  new_dataline->callback((fltk::Callback *)NewDataLineCB, this);
  datagroup = new fltk::ScrollGroup(10, 170, 580, 120, "");
  datagroup->scrollbar.linesize(20);
  newDataLineCB(widget, ptr);
  newDataLineCB(widget, ptr);
  ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
  ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
  ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
  ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
  /* buttons */
  fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
  cancel->image((fltk::Image *)cancel_icon);
  cancel->tooltip("Cancel");
  cancel->callback((fltk::Callback *)ResetBoundaryTabCB, this);
  fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Add");
  add->image(LOAD_ICON_IMAGE("images/add.png"));
  cancel->tooltip("Add the primitive");
  add->callback((fltk::Callback *)AddEllipseCB, this);
  boundary_subwindow->end();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 380);
}

void u2d_Umesh2dgui::ShowEllipseCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->showEllipseCB(widget, ptr);
}

void u2d_Umesh2dgui::showDataCB(fltk::Widget *widget, void *ptr) {
    /* POPULATE THE BOUNDARY_SUBWINDOW */
    resetBoundaryTabCB(widget, ptr);
    dataitem->box(fltk::DOWN_BOX);
    boundary_subwindow->begin();

    /* rotation */
    fltk::ValueInput *rotation = new fltk::ValueInput(190, 10, 100, 20, "Rotation angle: ");
    rotation->range(-360, 360);
    rotation->linesize(1.0);
    rotation->step(1);
    rotation->set_value(0);
    inputindex[0] = boundary_subwindow->find(rotation);

    /* name */
    fltk::Input *primname = new fltk::Input(450, 10, 100, 20, "Name (Optional): ");
    inputindex[1] = boundary_subwindow->find(primname);


    /* start point bordered box */
    fltk::Group *scontour = new fltk::Group(50, 31, 240, 47, "Displacement ");
    scontour->box(fltk::BORDER_FRAME);
    scontour->align(fltk::ALIGN_INSIDE_LEFT);

    /* trasx */
    fltk::ValueInput *trasx = new fltk::ValueInput(190, 34, 98, 20, "x: ");
    trasx->range(-1000, 1000);
    trasx->step(0.01);
    trasx->linesize(0.01);
    inputindex[2] = boundary_subwindow->find(trasx);

    /* trasy */
    fltk::ValueInput *trasy = new fltk::ValueInput(190, 55, 98, 20, "y: ");
    trasy->range(-1000, 1000);
    trasy->step(0.01);
    trasy->linesize(0.01);
    inputindex[3] = boundary_subwindow->find(trasy);

    /* data filename */
    fltk::Input *datafilename = new fltk::Input(COLUMN2, 80, 250, 20, "Filename :");
    inputindex[4] = boundary_subwindow->find(datafilename);

    /* Datagroup for [i|s|h]data */
    fltk::Button *new_dataline = new fltk::Button(10, 120, 580, 20, "Add data entry");
    new_dataline->callback((fltk::Callback *)NewDataLineCB, this);
    datagroup = new fltk::ScrollGroup(10, 150, 580, 140, "");
    datagroup->scrollbar.linesize(20);
    newDataLineCB(widget, ptr);
    newDataLineCB(widget, ptr);
    ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
    ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
    ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
    ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);

    /* Buttons (Add, Cancel)*/
    fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
    cancel->image((fltk::Image *)cancel_icon);
    cancel->tooltip("Cancel");
    cancel->callback((fltk::Callback *)ResetBoundaryTabCB, this);
    fltk::Button *add = new fltk::Button(320, 305, 100, 25, "Load");
    add->image(LOAD_ICON_IMAGE("images/add.png"));
    cancel->tooltip("Add the primitive");
    add->callback((fltk::Callback *)AddDataCB, this);
    boundary_subwindow->end();
    boundary_subwindow->redraw();
}

void u2d_Umesh2dgui::ShowDataCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->showDataCB(widget, ptr);
}

void u2d_Umesh2dgui::showRectangleCB(fltk::Widget *widget, void *ptr) {
    /* POPULATE THE BOUNDARY_SUBWINDOW */
    resetBoundaryTabCB(widget, ptr);
    rectangleitem->box(fltk::DOWN_BOX);
    boundary_subwindow->begin();

    /* spline*/
    fltk::ValueInput *splinepoints = new fltk::ValueInput(COLUMN2, 10, 100, 20, "N. points for spline aprox.: ");
    splinepoints->range(0, 10000);
    splinepoints->linesize(1.0);
    splinepoints->step(1);
    splinepoints->set_value(100);
    inputindex[0] = boundary_subwindow->find(splinepoints);

    /* name */
    fltk::Input *primname = new fltk::Input(COLUMN4, 10, 100, 20, "Name (Optional): ");
    inputindex[1] = boundary_subwindow->find(primname);

    /* start point bordered box */
    fltk::Group *scontour = new fltk::Group(COLUMN1, 31, 240, 47, "Bottom left point");
    scontour->box(fltk::BORDER_FRAME);
    scontour->align(fltk::ALIGN_INSIDE_LEFT);

    /* x1 */
    fltk::ValueInput *x1 = new fltk::ValueInput(COLUMN2, 34, 98, 20, "x: ");
    x1->range(-1000, 1000);
    x1->step(0.01);
    x1->linesize(0.01);
    inputindex[2] = boundary_subwindow->find(x1);

    /* y1 */
    fltk::ValueInput *y1 = new fltk::ValueInput(COLUMN2, 55, 98, 20, "y: ");
    y1->range(-1000, 1000);
    y1->step(0.01);
    y1->linesize(0.01);
    inputindex[3] = boundary_subwindow->find(y1);


    /* end point bordered box */
    fltk::Group *econtour = new fltk::Group(COLUMN3, 31, 240, 47, "Top right point");
    econtour->box(fltk::BORDER_FRAME);
    econtour->align(fltk::ALIGN_INSIDE_LEFT);

    /* x2 */
    fltk::ValueInput *x2 = new fltk::ValueInput(COLUMN4, 34, 98, 20, "x: ");
    x2->range(-1000, 1000);
    x2->step(0.01);
    x2->linesize(0.01);
    inputindex[4] = boundary_subwindow->find(x2);

    /* y2 */
    fltk::ValueInput *y2 = new fltk::ValueInput(COLUMN4, 55, 98, 20, "y: ");
    y2->range(-1000, 1000);
    y2->step(0.01);
    y2->linesize(0.01);
    inputindex[5] = boundary_subwindow->find(y2);

    /* Datagroup for [i|s|h]data */
    /*  fltk::Group *databox = new fltk::Group(10, 130, 580, 160, "");
    databox->box(fltk::BORDER_FRAME);
     */
    fltk::Button *new_dataline = new fltk::Button(10, 100, 580, 20, "Add data entry");
    new_dataline->callback((fltk::Callback *)NewDataLineCB, this);
    //  datagroup = new fltk::ScrollGroup(10, 140, 580, 160, "Insert [i|s|h]data here:");
    datagroup = new fltk::ScrollGroup(10, 130, 580, 160, "");
    datagroup->scrollbar.linesize(20);
    newDataLineCB(widget, ptr);
    newDataLineCB(widget, ptr);
    ((fltk::ValueInput *)(datagroup->child(1)))->set_value(0);
    ((fltk::ValueInput *)(datagroup->child(2)))->set_value(0.05);
    ((fltk::ValueInput *)(datagroup->child(4)))->set_value(1);
    ((fltk::ValueInput *)(datagroup->child(5)))->set_value(0.05);
    //  datagroup->box(fltk::BORDER_FRAME);
    //  datagroup->begin();
    //  fltk::Widget *box2 = new fltk::Widget(0, 0, 620, 160, "TEXT"); // this is to see the space avaiable
    //  fltk::Scroll *datascroll = new fltk::Scroll(600, 0, 20, 160, 0);
    //  datagroup->end();

    /* Buttons (Add, Cancel)*/
    fltk::Button *cancel = new fltk::Button(200, 305, 100, 25, "Cancel");
    cancel->image((fltk::Image *)cancel_icon);
    cancel->tooltip("Cancel");
    cancel->callback((fltk::Callback *)ResetBoundaryTabCB, this);
    fltk::Button *add = new fltk::Button(320, 305, 100, 25, "&Add");
    add->image(LOAD_ICON_IMAGE("images/add.png"));
    cancel->tooltip("Add the primitive");
    add->callback((fltk::Callback *)AddRectangleCB, this);
    boundary_subwindow->end();
    boundary_subwindow->redraw();
}

void u2d_Umesh2dgui::ShowRectangleCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->showRectangleCB(widget, ptr);
}

void u2d_Umesh2dgui::newDataLineCB(fltk::Widget *widget, void *ptr) {
  int currentpos;
  datagroup->begin();
  /* This is a hard hack since inputs (and any other thing) got created
   * from actual origin and not from widget (datagroup in this case)
   * absolute position, anyway the first time the input got creted with
   * the bar fully scored an unexpected 10 pixel offset is generated.
   * This can be an initialization bug in fltk code.
   * I figured out the problem and resoleved it, it is a bad, but not
   * buggy, behavor or fltk. It just wipesout white space if not filled
   * by some widget (window or anything) */
  /*
  if(datagroup->yposition() < 0) {
    currentpos = ydataposition;
  }
  else {
    currentpos = ydataposition - datagroup->yposition();
  }
  */
  datagroup->yposition() < 0 ?\
    currentpos = ydataposition :\
    currentpos = ydataposition - datagroup->yposition();
  //  fprintf(stdout, "yposition(): %i\n", datagroup->yposition());
  //  fprintf(stdout, "currentpos: %i\n", currentpos);
  //  fflush(stdout);
  //  ydataposition = ydataposition - datagroup->yposition();
  fltk::ValueInput *idata = new fltk::ValueInput(50, currentpos, 100, 20, "idata: ");
  idata->range(0, 3);
  idata->linesize(1.0);
  idata->set_value(1);
  fltk::ValueInput *sdata = new fltk::ValueInput(250, currentpos, 100, 20, "sdata: ");
  sdata->range(0, 1);
  sdata->linesize(0.01);
  sdata->set_value(1);
  fltk::ValueInput *hdata = new fltk::ValueInput(450, currentpos, 100, 20, "hdata: ");
  hdata->range(0, 1);
  hdata->linesize(0.01);
  hdata->set_value(0.05);
  datagroup->end();
  datagroup->redraw();
  //  datagroup->redraw(0, currentpos, 620, 20);  // maybe this is not anymore necessary, but should be correct.
  ydataposition = ydataposition + 20;

}

void u2d_Umesh2dgui::NewDataLineCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->newDataLineCB(widget, ptr);
}

void u2d_Umesh2dgui::resetBoundaryTabCB(fltk::Widget *widget, void *ptr) {
  /* reset the scroll position */
  ydataposition = 0;
  /* clean the index for primitive values */
  for(int i = 0; i < 10; i++) inputindex[i] = 0;
  /* reset button state */
  lineitem->box(fltk::NO_BOX);
  circleitem->box(fltk::NO_BOX);
  ellipseitem->box(fltk::NO_BOX);
  dataitem->box(fltk::NO_BOX);
  rectangleitem->box(fltk::NO_BOX);
  /* clear the window */
  boundary_subwindow->clear();
  boundary_subwindow->redraw();
  //  boundary_subwindow->redraw(0, 0, 620, 330);
}

void u2d_Umesh2dgui::ResetBoundaryTabCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->resetBoundaryTabCB(widget, ptr);
}

void u2d_Umesh2dgui::addLineCB(fltk::Widget *widget, void *ptr) {
  int ndata;
  int validdata = 0;
  Line *prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->x1 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->y1 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->x2 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->y2 = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  std::cout << "--== Adding line primitive ==--" << std::endl;
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arrangements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);
  std::cout << "--== Done ==--" << std::endl;
}

void u2d_Umesh2dgui::AddLineCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addLineCB(widget, ptr);
}

void u2d_Umesh2dgui::addCircleCB(fltk::Widget *widget, void *ptr) {
  int ndata;
  int validdata = 0;
  Circle *prim = (Circle *)(u2d_core->add_prim(PR_CIRCLE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->radius = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->centerx = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->centery = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->startarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  prim->endarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),/* spline */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* radius */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* centerx */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(), /* centery */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(), /* startarc */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value()); /* endarc */
  std::cout << "--== Adding circle primitive ==--" << std::endl;
  printf("Spline: %i radius: %f centerx: %f centery: %f startarc: %f endarc: %f\n", prim->spline, prim->radius, prim->centerx, prim->centery, prim->startarc, prim->endarc);
  fflush(stdout);
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This for cycle could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);
  std::cout << "--== Done ==--" << std::endl;
}

void u2d_Umesh2dgui::AddCircleCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addCircleCB(widget, ptr);
}

void u2d_Umesh2dgui::addEllipseCB(fltk::Widget *widget, void *ptr) {
  int ndata;
  int validdata = 0;
  Ellipse *prim = (Ellipse *)(u2d_core->add_prim(PR_ELLIPSE));
#ifdef DEBUGGING
  printf("boundary_subwindow ha %i figli\n", boundary_subwindow->children());
#endif
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  /* Changed to use the new function
  prim->spline = (int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value();
  prim->semix = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value();
  prim->semiy = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value();
  prim->centerx = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value();
  prim->centery = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value();
  prim->startarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value();
  prim->endarc = ((fltk::ValueInput *)boundary_subwindow->child(inputindex[7]))->value();
  */
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),  /* spline */ \
  	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* semix */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* semiy */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(), /* centerx */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(), /* centery */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[6]))->value(), /* startarc */ \
    	  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[7]))->value()); /* endarc */
  std::cout << "--== Adding ellipse primitive ==--" << std::endl;
  printf("Spline: %i semix: %f semiy: %f centerx: %f centery: %f startarc: %f endarc: %f\n", prim->spline, prim->semix, prim->semiy, prim->centerx, prim->centery, prim->startarc, prim->endarc);
  fflush(stdout);
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);
  std::cout << "--== Done ==--" << std::endl;
}

void u2d_Umesh2dgui::AddEllipseCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addEllipseCB(widget, ptr);
}

void u2d_Umesh2dgui::addDataCB(fltk::Widget *widget, void *ptr) {
  int ndata;
  int validdata = 0;
  int retval;
  int path_max;
  
  std::cout << "--== Adding data primitive ==--" << std::endl;
  
  Data *prim = (Data *)(u2d_core->add_prim(PR_DATA));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);

  /* Expand filename */
#ifdef PATH_MAX
  path_max = PATH_MAX;
#else
  path_max = pathconf (((fltk::Input *)boundary_subwindow->child(inputindex[4]))->value(), _PC_PATH_MAX);
  if (path_max <= 0)
    path_max = 4096;
#endif
  char *filename_expanded = NULL;
  filename_expanded = (char *)malloc(path_max * sizeof(char));
  if(filename_expanded == NULL){
    fprintf(stderr, "Cannot allocate memory for pathname expansion\n");
    outputMessage("ERROR: Cannot allocate memory for pathname expansion");
    u2d_core->del_prim(prim);
    return;
  }
  memset(filename_expanded, 0, path_max);
  realpath(((fltk::Input *)boundary_subwindow->child(inputindex[4]))->value(),\
	   filename_expanded);

  retval = prim->set_coord(filename_expanded,  /* filename */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(), /* trasx */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(), /* trasy */ \
			   ((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value()); /* rotation */
  free(filename_expanded);
  if(retval < 0) {
    outputMessage("ERROR loading data file");
    if(retval == NO_FILE) fltk::alert("Error opening file: Invalid data filename\n");
    else if(retval == INVALID_FILE) fltk::alert("Error reading file: check file syntax\n");
    else if(retval == NOT_A_FILE) fltk::alert("Error with file: not a regular file\n");
    else if(retval == NO_EXIST) fltk::alert("Error with filename: check filename syntax\n");
    else fltk::alert("Error opening file: Unknown reason\n");
    u2d_core->del_prim(prim);
    return;
  }
  outputMessage("Data file loaded, acquired %i points", prim->npoints);
  printf("Filename: %s trasx: %f trasy: %f rotation: %f\n",\
	 prim->filename, prim->trasx, prim->trasy,\
	 prim->rotation);
  fflush(stdout);
  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* This could be done nicer with k%3 as second argument and some
   * other arrangements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);
  std::cout << "--== Done ==--" << std::endl;
}

void u2d_Umesh2dgui::AddDataCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addDataCB(widget, ptr);
}

void u2d_Umesh2dgui::addRectangleCB(fltk::Widget *widget, void *ptr) {
/* Bottom left -> Top left */
  int ndata;
  int validdata = 0;
  Line *prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  std::cout << "--== Adding rectangle primitive ==--" << std::endl;
  std::cout << "Spline: " << prim->spline
          << " x1: " << prim->x1 << " y1: " << prim->y1
          << " x2: " << prim->x2 << " y2: " << prim->y2
          << std::endl;
  fflush(stdout);

  ndata = datagroup->children();
  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);

/* Top left -> Top right */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);

/* Top right -> Bottom right */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[5]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);

/* Bottom right -> Bottom left */
  validdata = 0;
  prim = (Line *)(u2d_core->add_prim(PR_LINE));
  strncpy(prim->primname, ((fltk::Input *)boundary_subwindow->child(inputindex[1]))->value(), 32);
  prim->set_coord((int)((fltk::ValueInput *)boundary_subwindow->child(inputindex[0]))->value(),\
                  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[4]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[2]))->value(),\
		  ((fltk::ValueInput *)boundary_subwindow->child(inputindex[3]))->value());
  printf("Spline: %i x1: %f y1: %f x2: %f y2: %f\n", prim->spline, prim->x1, prim->y1, prim->x2, prim->y2);
  fflush(stdout);

  prim->allocatedata(ndata/3);
  /* this could be done nicer with k%3 as second argument and some
   * other arragnements, but it's ok for now
   */
  for(int k = 0; k < ndata; k = k + 3) {
    if(((fltk::ValueInput *)datagroup->child(k))->value() > 0) {
#ifdef DEBUGGING
      fprintf(stdout, "k/3: %i child n.: %i valid data (x): %i\n", k/3, k, validdata);
      fflush(stdout);
#endif
      prim->setdata(validdata, 0, ((fltk::ValueInput *)datagroup->child(k))->value());
      prim->setdata(validdata, 1, ((fltk::ValueInput *)datagroup->child(k + 1))->value());
      prim->setdata(validdata, 2, ((fltk::ValueInput *)datagroup->child(k + 2))->value());
      validdata++;
      prim->curdata++;
    }
  }
  addPrimitive(prim);
  std::cout << "--== Done ==--" << std::endl;
}

void u2d_Umesh2dgui::AddRectangleCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addRectangleCB(widget, ptr);
}

void u2d_Umesh2dgui::addPrimitive(Primitive *prim) {
  u2d_gl->set_prim_list_start(u2d_core->get_prim_list_start(), \
		              u2d_core->get_prim_number());
  u2d_gl->guess_ref_size(prim);
  u2d_gl->redraw();
  prim_browser->add(u2d_core->get_prim_string(prim), 0, NULL, 0, 0);

}

void u2d_Umesh2dgui::openItemCB(fltk::Widget *widget, void *ptr) {
    fltk::message("Open Feature not avaialbe yet\n");
}

void u2d_Umesh2dgui::saveItemCB(fltk::Widget *widget, void *ptr) {
  if(strlen(((fltk::Input *)creation_subwindow->child(filenameinput_i))->value()) < 1) saveAsItemCB(widget, ptr);
  else creationNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::saveAsItemCB(fltk::Widget *widget, void *ptr) {
  openFileBrowserCB(widget, ptr);
  creationNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::quitItemCB(fltk::Widget *widget, void *ptr) {
  window->hide();
  opengl_window->hide();
  prim_browser_window->hide();
  /* calling the upper 2 or this one is exactly the same, no idea what is better */
  /* exit(0); */
}

void u2d_Umesh2dgui::aboutItemCB(fltk::Widget *widget, void *ptr) {
    fltk::message("%s by Gabriele Dini Ciacci - version %s\n", PACKAGE, VERSION);
}


void u2d_Umesh2dgui::OpenItemCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->openItemCB(widget, ptr);
}

void u2d_Umesh2dgui::SaveItemCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->saveItemCB(widget, ptr);
}

void u2d_Umesh2dgui::SaveAsItemCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->saveAsItemCB(widget, ptr);
}

void u2d_Umesh2dgui::QuitItemCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->quitItemCB(widget, ptr);
}

void u2d_Umesh2dgui::AboutItemCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->aboutItemCB(widget, ptr);
}

void u2d_Umesh2dgui::boundaryTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    boundarygroup->hide();
    domaingroup->show();
    outputMessage("Define domain constraints");
}

void u2d_Umesh2dgui::domainTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    domaingroup->hide();
    boundarygroup->show();
    outputMessage("Back to boundary definition");
}

void u2d_Umesh2dgui::domainTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    domaingroup->hide();
    creationgroup->show();
    outputMessage("Define output options");
}

void u2d_Umesh2dgui::creationTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    creationgroup->hide();
    domaingroup->show();
    outputMessage("Back to domain definition");
}

void u2d_Umesh2dgui::sidesTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    /* Check if boundary conditions have been set */
    if(sidesradiogroup.empty()) {
        fltk::alert("Please load the mesh and set the boundary condition type in the \"Sides\" tab first.");
        outputMessage("Please load the mesh and set the boundary condition type in the \"Sides\" tab first.");
        return;
    }
    sidesgroup->hide();
    parametersgroup->show();
    outputMessage("Define problem parameters");
}


void u2d_Umesh2dgui::parametersTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    parametersgroup->hide();
    sidesgroup->show();
    outputMessage("Back to boundary condition type definition");
}

void u2d_Umesh2dgui::setUmesh2dFormatCB(fltk::Widget *widget, void *ptr) {
      fileformat = FORMAT_UMESH2D;
}
void u2d_Umesh2dgui::setUmesh2dFEFormatCB(fltk::Widget *widget, void *ptr) {
    fileformat = FORMAT_UMESH2D_FE;
}
void u2d_Umesh2dgui::setUmesh2dFVFormatCB(fltk::Widget *widget, void *ptr) {
   fileformat = FORMAT_UMESH2D_FV;
}

void u2d_Umesh2dgui::BoundaryTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->boundaryTabNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::DomainTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->domainTabNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::DomainTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->domainTabPrevButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::SidesTabNextButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->sidesTabNextButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::ParametersTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->parametersTabPrevButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::CreationTabPrevButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->creationTabPrevButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::SetUmesh2dFormatCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->setUmesh2dFormatCB(widget, ptr);
}

void u2d_Umesh2dgui::SetUmesh2dFEFormatCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->setUmesh2dFEFormatCB(widget, ptr);
}

void u2d_Umesh2dgui::SetUmesh2dFVFormatCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->setUmesh2dFVFormatCB(widget, ptr);
}

void u2d_Umesh2dgui::glViewItemCB(fltk::Widget *widget, void *ptr) {
  if(opengl_window->visible()){
    opengl_window->hide();
  }
  else {
    opengl_window->show();
    opengl_window->position(window->x() + window->w() + MANAGER_BORDER, window->y());
  }
}

void u2d_Umesh2dgui::GlViewItemCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->glViewItemCB(widget, ptr);
}

void u2d_Umesh2dgui::browserViewItemCB(fltk::Widget *widget, void *ptr) {
    if (prim_browser_window->visible()) {
        prim_browser_window->hide();
    } else {
        prim_browser_window->show();
        prim_browser_window->position(window->x(), window->y() - BROWSER_H - TITLEBAR_OFFSET);
    }
}

void u2d_Umesh2dgui::BrowserViewItemCB(fltk::Widget *widget, void *ptr){
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->browserViewItemCB(widget, ptr);
}

void u2d_Umesh2dgui::deletePrimitiveFromBrowserButtonCB(fltk::Widget *widget, void *ptr) {
  fltk::Widget* current = prim_browser->goto_top();
  while (current) {
    if (current->selected()) {
      /* This work in case of hierarchical menus too.
       * In my case the current->parent() should always be
       * prim_browser */
      fltk::Group* parent = current->parent();
      parent->remove(current);
      /* Now we delete the primitive from the list like:
       * id = current->value() to get the removed primitive id
       * del_prim(get_prim_by_id(id)); */
      u2d_core->del_prim(u2d_core->get_prim_by_label(current->label()));
      delete current;
      parent->relayout();
      /* We start again from top to see if there are
       * still other primitives selected */
      current = prim_browser->goto_top();
    }
    else {
      current = prim_browser->next();
    }
  }

  /* We have to update the openGL window too */
  u2d_gl->set_prim_list_start(u2d_core->get_prim_list_start(), \
			      u2d_core->get_prim_number());
  /* We have to recalculate the reference sizes from
   * scratch... VERY costly */
  u2d_gl->guess_ref_size();
  /* We can finally redraw */
  u2d_gl->redraw();
}

void u2d_Umesh2dgui::DeletePrimitiveFromBrowserButtonCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->deletePrimitiveFromBrowserButtonCB(widget, ptr);
}

void u2d_Umesh2dgui::runMainCoreCB(fltk::Widget *widget, void *ptr) {
    /* Calculate closed objects */
    fprintf(stdout, "Calculating closed objects and vertex enumeration...\n");
    u2d_core->find_objects();
    fprintf(stdout, "Done!\n");
}

void u2d_Umesh2dgui::RunMainCoreCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->runMainCoreCB(widget, ptr);
}

void u2d_Umesh2dgui::outputMessage(char *format, ...) {
  va_list ap;
  va_start(ap, format);
  int length = vsnprintf(msgbuf, MSGBUF_LENGTH, format, ap);
  memset(msgbuf + sizeof(char)*length + 1, 0, MSGBUF_LENGTH - sizeof(char)*length - 1);
  msgbar->value(msgbuf);
  va_end(ap);
}

void u2d_Umesh2dgui::addBackgridLineCB(fltk::Widget *widget, void *ptr) {
    int currentpos;
    backgridgroup->begin();
    /* This is a hard hack since inputs (and any other thing) got created
     * from actual origin and not from widget (backgrid in this case)
     * absolute position, anyway the first time the input got created with
     * the bar fully scored and unexpected 10 pixel offset is generated.
     * I suspect this is an initialization bug in fltk code.
     * I figured out the problem and resolved it, it is a bad, but not
     * buggy, behavior or fltk. It just wipes out white space if not filled
     * by some widget (window or anything) */

    backgridgroup->yposition() < 0 ?  \
    currentpos = ybackgridposition :  \
 currentpos = ybackgridposition - backgridgroup->yposition();

    fltk::ValueInput *x = new fltk::ValueInput(40, currentpos, 100, 20, "x: ");
    x->linesize(0.01);
    fltk::ValueInput *y = new fltk::ValueInput(180, currentpos, 100, 20, "y: ");
    y->linesize(0.01);
    fltk::ValueInput *hdata = new fltk::ValueInput(330, currentpos, 100, 20, "hdata: ");
    hdata->linesize(0.01);
    fltk::Button *backgrid_del = new fltk::Button(440, currentpos, 80, 20, "delete");
    backgrid_del->callback((fltk::Callback *)DeleteBackgridLineCB, this);
    backgridgroup->end();
    backgridgroup->redraw();
    ybackgridposition = ybackgridposition + 20;
}

void u2d_Umesh2dgui::AddBackgridLineCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addBackgridLineCB(widget, ptr);
}

void u2d_Umesh2dgui::addSteinerLineCB(fltk::Widget *widget, void *ptr) {
  int currentpos;
  steinergroup->begin();
  /* This is a hard hack since inputs (and any other thing) got created
   * from actual origin and not from widget (backgrid in this case)
   * absolute position, anyway the first time the input got created with
   * the bar fully scored and unexpected 10 pixel offset is generated.
   * I suspect this is an initialization bug in fltk code.
   * I figured out the problem and resolved it, it is a bad, but not
   * buggy, behavior or fltk. It just wipes out white space if not filled
   * by some widget (window or anything) */

  steinergroup->yposition() < 0 ?\
    currentpos = ysteinerposition :\
    currentpos = ysteinerposition - steinergroup->yposition();

  fltk::ValueInput *x = new fltk::ValueInput(40, currentpos, 100, 20, "x: ");
  x->linesize(0.01);
  fltk::ValueInput *y = new fltk::ValueInput(240, currentpos, 100, 20, "y: ");
  y->linesize(0.01);
  fltk::Button *steiner_del = new fltk::Button (440, currentpos, 80, 20, "delete");
  steiner_del->callback((fltk::Callback *)DeleteSteinerLineCB, this);
  steinergroup->end();
  steinergroup->redraw();
  ysteinerposition = ysteinerposition + 20;
}

void u2d_Umesh2dgui::AddSteinerLineCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->addSteinerLineCB(widget, ptr);
}

void u2d_Umesh2dgui::deleteBackgridLineCB(fltk::Widget *button, void *ptr) {
  /* An integer that will contain the child id of the button */
  int button_line = 0;

  /* We get the parent, should be the ScrollGroup */
  fltk::Group *parent = ((fltk::Button *)button)->parent();

  /* we find the button child index */
  button_line = parent->find(button);

  /* We delete the lines, notice that we delete for 3 times the same index.
   * This is because the children get renumbered after we delete one
   * so if we delete child n the n+1 child becomes himself the n child.
   */
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
  delete parent->child(button_line - 3);
}

void u2d_Umesh2dgui::DeleteBackgridLineCB(fltk::Widget *button, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->deleteBackgridLineCB(button, NULL);
}

void u2d_Umesh2dgui::deleteSteinerLineCB(fltk::Widget *button, void *ptr) {
  /* an integer that will contain the child id of the button */
  int button_line = 0;

  /* we get the parent, should be the ScrollGroup */
  fltk::Group *parent = ((fltk::Button *)button)->parent();

  /* we find the button child index */
  button_line = parent->find(button);

  /* We delete the lines, notice that we delete for 3 times the same index.
   * This is because the children get renumbered after we delete one
   * so if we delete child n the n+1 child becomes himself the n child.
   */
  delete parent->child(button_line - 2);
  delete parent->child(button_line - 2);
  delete parent->child(button_line - 2);
}

void u2d_Umesh2dgui::DeleteSteinerLineCB(fltk::Widget *button, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->deleteSteinerLineCB(button, NULL);
}

void u2d_Umesh2dgui::solveProblemCB(fltk::Widget *widget, void *ptr) {    
    /* Check if boundary conditions have been set */
    if(sidesradiogroup.empty()) {
        fltk::alert("Please load the mesh and set the boundary condition type in the \"Sides\" tab first.");
        outputMessage("Please load the mesh and set the boundary condition type in the \"Sides\" tab first.");
        return;
    }

    helmholtz = new u2d_Helmholtz(this->project->getProjectAbsoluteDir(), this->project->getProjectName());

    std::cout << "--== Setting boundary condition types on sides ==--" << std::endl;
    int sideid = 1;
    for (std::vector<fltk::Group *>::iterator it = sidesradiogroup.begin();
            it != sidesradiogroup.end(); it++) {
        fltk::RadioButton *button = (fltk::RadioButton *)(*it)->child(0);
        if (button->value()) {
            // dirichlet
            helmholtz->setSideDirichlet(sideid);
            std::cout << "Dirichlet: " << sideid << std::endl;
        } else {
            //neumann
            helmholtz->setSideNeumann(sideid);
            std::cout << "Neumann: " << sideid << std::endl;
        }
        sideid++;
    }
    std::cout << "--== Done ==--" << std::endl;

    /* Convert the values to doubles */
    helmholtz->setGamma(strtod(gammainput->value(), NULL));
    helmholtz->setAlpha(strtod(alphainput->value(), NULL));

    /* Check if file or equation was provided and and set the vectors */
    try {
        /* a(s) */
        if (((fltk::RadioButton *)(this->vectorradiogroup[0]->child(0)))->value()) {
            helmholtz->setAVector(string(avectorinput->value()));
        } else {
            u2d_LoadScalarField loadavector(avectorinput->value(), false);
            helmholtz->setAVector(loadavector.getVector());
        }
        /* b(s) */
        if (((fltk::RadioButton *)(this->vectorradiogroup[1]->child(0)))->value()) {
            helmholtz->setBVector(string(bvectorinput->value()));
        } else {
            u2d_LoadScalarField loadbvector(bvectorinput->value(), false);
            helmholtz->setBVector(loadbvector.getVector());
        }
        /* b(s) Field */
        if (((fltk::RadioButton *)(this->vectorradiogroup[2]->child(0)))->value()) {
            helmholtz->setBField(string(bfieldinput->value()), string(bfieldinput->value())); // TODO The are the same for x and y just now
        } else {
            u2d_LoadVectorField loadbfield(bfieldinput->value(), 2, false);
            helmholtz->setBField(loadbfield.getField());
        }
        /* source(r) */
        if (((fltk::RadioButton *)(this->vectorradiogroup[3]->child(0)))->value()) {
            helmholtz->setSource(string(sourcevectorinput->value()));
        } else {
            u2d_LoadScalarField loadsource(sourcevectorinput->value(), false);
            helmholtz->setSource(loadsource.getVector());
        }
        /* k(r) */
        if (((fltk::RadioButton *)(this->vectorradiogroup[4]->child(0)))->value()) {
            helmholtz->setKVector(string(kvectorinput->value()));
        } else {
            u2d_LoadScalarField loadkvector(kvectorinput->value(), false);
            helmholtz->setKVector(loadkvector.getVector());
        }
        /* v(r) */
        if (((fltk::RadioButton *)(this->vectorradiogroup[5]->child(0)))->value()) {
            helmholtz->setVField(string(vxfieldinput->value()), string(vyfieldinput->value()));
        } else {
            u2d_LoadVectorField loadvfield(vxfieldinput->value(), 2, false);
            helmholtz->setVField(loadvfield.getField());
        }
    } catch (exception& e) {
        std::cout << e.what() << std::endl;
    } catch (exception *e) {
        std::cout << e->what() << std::endl;
    }
    
    u2d_Problem *problem = helmholtz->run();
       
    /* Output to console */
    //std::cout << solver.getSolution() << std::endl;

    /* Output to mesh window */
    u2d_glmesh->clearDisplayObjects();
    u2d_glmesh->addDisplayObject(u2d_DisplaySet::FromMeshDomainWithSolution(this->mesh, problem));
    u2d_glmesh->addDisplayObject(u2d_DisplaySet::CreateKey(this->mesh, problem));
    
    /* Switch to solution display tab */
    parametersgroup->hide();
    meshgroup->show();
    
    /* Output to plotmtv format */
    u2d_PlotmtvOutput plotmtvout(this->project, this->mesh, problem);
    plotmtvout.output();
    
    
    
}

void u2d_Umesh2dgui::SolveProblemCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->solveProblemCB(widget, ptr);
}

void u2d_Umesh2dgui::loadMeshCB(fltk::Widget *widget, void *ptr) {
    /* Clean the eventually old loaded mesh values */
    /* Delete the widgets */
    for (std::vector<fltk::Group *>::iterator it = sidesradiogroup.begin();
            it != sidesradiogroup.end(); it++) {
        delete *it;
    }
    for (std::vector<std::stringstream *>::iterator it = sidesradiogroup_labels.begin();
            it != sidesradiogroup_labels.end(); it++) {
        delete *it;
    }
    /* Clear the vectors holding the pointers */
    sidesradiogroup.clear();
    sidesradiogroup_labels.clear();

    if(project == NULL) {
        fltk::alert("No project loaded. Load one first.");
        outputMessage("Load a project first!");
        return;
    }

    try {
        delete mesh;
        u2d_Load_Umesh2d mesh_loader(this->project->getProjectAbsoluteDir(), this->project->getProjectName(), true);
        this->mesh = mesh_loader.getMesh();
    } catch (exception &e) {
        std::cout << e.what() << std::endl;
        fltk::alert("Mesh loading failed.");
        outputMessage("Mesh loading failed: %s", e.what());
        return;
    }   
    

    std::stringstream out;
    out << this->mesh->getDomainNodeNumber();
    domainnodesoutput->value(out.str().c_str());

    std::stringstream out2;
    out2 << this->mesh->getBoundaryNodeNumber();
    boundarynodesoutput->value(out2.str().c_str());

    std::stringstream out3;
    out3 << this->mesh->getDomainElementNumber();
    domainelementsoutput->value(out3.str().c_str());

    std::stringstream out4;
    out4 << this->mesh->getBoundaryElementNumber();
    boundaryelementsoutput->value(out4.str().c_str());

    std::vector<int> sides = this->mesh->getSideIdList();
    std::stringstream out5;
    for(std::vector<int>::iterator it = sides.begin(); it != sides.end(); it++){
        out5 << *it << " ";
    }
    sidesoutput->value(out5.str().c_str());

    outcontour_sidesinfo->begin();

    sideslistgroup = new fltk::ScrollGroup(50, 1, 449, 228, "");
    sideslistgroup->scrollbar.linesize(20);

    sideslistgroup->box(fltk::BORDER_FRAME);
    
    /* boundary sides radiobuttons list */
    int idx = 0;
    for (std::vector<int>::iterator it = sides.begin(); it != sides.end(); it++) {
        /* Create and set the labels */
        std::stringstream *tmpout = new std::stringstream();
        (*tmpout) << "Side " << *it;
        sideslistgroup->begin();
        sidesradiogroup_labels.push_back(tmpout);
        fltk::Group *radiogroup = new fltk::Group(1, 20 * idx, 400, 20, "");
        radiogroup->align(fltk::ALIGN_INSIDE_LEFT);
        //radiogroup->box(fltk::BORDER_FRAME);
        radiogroup->copy_label((*tmpout).str().c_str());
        /* Create the radiobuttons */
        radiogroup->begin();
        fltk::RadioButton *isdirichletradio = new fltk::RadioButton(140, 0, 100, 20, "Dirichlet");
        isdirichletradio->value(true); // Set dirichlet as default
        fltk::RadioButton *isrobinradio = new fltk::RadioButton(260, 0, 100, 20, "Robin");
        //sides_checkbuttons.push_back(new fltk::CheckButton(340 + 30*idx, 32, 20, 20, (*tmpout).str().c_str()));
        radiogroup->end();
        sideslistgroup->end();
        sidesradiogroup.push_back(radiogroup);
        idx++;
    }
    outcontour_sidesinfo->end();
}

void u2d_Umesh2dgui::LoadMeshCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->loadMeshCB(widget, ptr);
}

void u2d_Umesh2dgui::runHelmholtzSquareTestCB(fltk::Widget *widget, void *ptr) {        
    string projectPath("../test/uhelm2d_square");
    string projectName("square_c");
    
    /* Delete current project and load a new one */
    delete this->project;
    this->project = u2d_Project::LoadProject(projectPath, projectName);

    /* Load the mesh for the current project */
    loadMeshCB(widget, ptr);

    /* Set the boundary condition as expected for Helmholtz square test problem */
    ((fltk::RadioButton *)(sidesradiogroup[3]->child(0)))->value(false);
    ((fltk::RadioButton *)(sidesradiogroup[3]->child(1)))->value(true);
    
    /* Set vector input type to file */
    for(std::vector<fltk::Group *>::iterator it = this->vectorradiogroup.begin();
            it != this->vectorradiogroup.end(); it++) {
        ((fltk::RadioButton *)((*it)->child(0)))->value(false);
        ((fltk::RadioButton *)((*it)->child(1)))->value(true);
    }
    
    /* Set input to test vectors*/
    gammainput->value("10.0");
    alphainput->value("2.0");

    /* source(r) */
    string sourceVectorPath = projectPath + string("/source");
    sourcevectorinput->value(sourceVectorPath.c_str());
    /* a(s) */
    string aVectorPath = projectPath + string("/avector");
    avectorinput->value(aVectorPath.c_str());
    /* b(s) */
    string bVectorPath = projectPath + string("/bvector");
    bvectorinput->value(bVectorPath.c_str());
    /* k(r) */
    string kVectorPath = projectPath + string("/kvector");
    kvectorinput->value(kVectorPath.c_str());
    /* b(s) */
    string bFieldPath = projectPath + string("/bfield");
    bfieldinput->value(bFieldPath.c_str());
    /* v(r) */
    string vFieldPath = projectPath + string("/vfield");
    vxfieldinput->value(vFieldPath.c_str());

    /* Solve the problem */
    solveProblemCB(widget, ptr);
}

void u2d_Umesh2dgui::RunHelmholtzSquareTestCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->runHelmholtzSquareTestCB(widget, ptr);
}

void u2d_Umesh2dgui::DebugCB(fltk::Widget *widget, void *ptr) {
    u2d_Umesh2dgui *umesh2dgui = reinterpret_cast<u2d_Umesh2dgui *>(ptr);
    umesh2dgui->debugCB(widget, ptr);
}

void u2d_Umesh2dgui::debugCB(fltk::Widget *widget, void * /*ptr*/) {
    boundarygroup->hide();
    meshgroup->show();

    //u2d_Load_PlotMtv dataloader("/home/gab/tmp/create1/gridplot.create1");
    //u2d_glmesh->clearDisplayObjects();
    //u2d_glmesh->addDisplayObject(dataloader.displaySet); 
    
    //string projectPath("/home/gab/tmp/create4");
    //string projectName("create4");
    
    /* Delete current project and load a new one */
    //delete this->project;
    //this->project = u2d_Project::LoadProject(projectPath, projectName);

    /* Load the mesh for the current project */
    //loadMeshCB(widget, NULL);
 
    //u2d_glmesh->clearDisplayObjects();
    //u2d_glmesh->addDisplayObject(u2d_DisplaySet::FromMeshDomain(this->mesh)); 
    
    
    //u2d_glmesh->addDisplayObject(new u2d_DisplayText("Text1"));
   

    
    
    //u2d_DisplaySet *testSet = new u2d_DisplaySet("TestSet");
    //testSet->pushVertexCoord(0,0);
    //testSet->pushVertexCoord(0,10);
    //testSet->pushVertexCoord(1,10);
    //testSet->pushVertexCoord(1,0);
    //testSet->setFillMode(GL_FILL)
    //        .setCoordPerVertex(2)
    //        .setVertexPerElement(3)
    //        .setVertexType(GL_QUADS)
    //        .setLineColor(1.0, 1.0, 1.0);
    //u2d_glmesh->addDisplayObject(testSet);
    
    for(int l = 1; l < 2; l++ ){
    u2d_DisplayText *text2 = new u2d_DisplayText("Text2");
    text2->setSize(0.01)
            .setPosition(3.0, 1.0*l, 0);
            (*text2) << "Very long string that i hope does not do strange stuff ";
            (*text2) << l;
    u2d_glmesh->addDisplayObject(text2);    
    }
    
    //u2d_glmesh->addDisplayObject(u2d_DisplaySet::CreateKey(this->mesh, NULL));
    
    
    
   
    
    
}

void debug_cb() {
    
    u2d_Helmholtz::SquareTest();
    
    
    /* Lniear color function test 
     * float colors[3];
    for(int i = 45; i < 34325; i += 300) {
        u2d_Math::LinearRGB(colors, i, 45, 34325);
        std::cout << "R: " << colors[0] << " G: " << colors[1] << " B: " << colors[2] << std::endl;
    }
     */
  

    
/*
    FILE *file = NULL;
    unsigned int npoints = 0;
    int val = 0;
    float *points;

    file = fopen("2dmesh", "r");
    if(file == NULL) {
        fprintf(stderr, "Cannot open file\n");
        return;
    }
    val = fscanf(file, "%d\n", &npoints);
    if(val != 1) {
        fprintf(stderr, "Cannot read the number of points in file\n");
        return;
    }
    points = (float *)calloc(npoints, 2*sizeof(float)); 
    for(unsigned int i = 1; i < npoints*2; i = i + 2) {
        val = fscanf(file, "  %f %f       0\n", (points + i - 1), (points + i));
        if(val != 2) {
            fprintf(stderr, "debug_cb: Cannot acquire correctly from"\
                    "file 2dmesh. Check line %i "\
                    "is in the correct format i.e. 0.034 0.23\n",\
                    i);
            fclose(file);
            return;
        }
        printf("Acquired point: (%f, %f)\n", *(points + i - 1), *(points + i));
    }
    fclose(file);

    Display_Set display_set;
    display_set.vertex_array = points;
    display_set.vertex_array_n = npoints;
    display_set.type = GL_LINES;
*/
//2>&1

    //create_dir();
}

void u2d_Umesh2dgui::setProject(u2d_Project *project_) {
    if(this->project != NULL) {
        delete this->project;
        this->project = NULL;
    }
    this->project = project_;
}

void u2d_Umesh2dgui::executeMesher(u2d_Project *project) {
    /* Old command line for old umesh2d:
     * string cmd_line = "echo " + u2d_umesh2d_output->getFilename() + " | ./umesh2dmesher";
     */
    string mesher_path("umesh2d/umesh2d");
    string cmd_line = mesher_path + " " + project->getProjectName() + " " + project->getProjectAbsoluteDir();

    FILE *fp = popen(cmd_line.c_str(), "r");
    setvbuf(fp, NULL, _IOLBF, 256);

    char buffer[256];
    memset(buffer, 0, 256);
    while(fgets(buffer, 256, fp)) {
        printf("%s", buffer);
        shell_output->append(buffer);
    }
    pclose(fp);
}

void u2d_Umesh2dgui::drawMesh(u2d_Project *project) {
    u2d_Load_PlotMtv dataloader(project->getMeshGridplotFilename().c_str());
    u2d_glmesh->clearDisplayObjects();
    u2d_glmesh->addDisplayObject(dataloader.displaySet);
}


/***************** TO CHECK AND REMOVE *******************/

/* BUG we have  no way to know if we are in the correct dir if
 * the user name the last dir the same project dir like:
 * /home/name/mesh/mesh
 * This will cause confusion as if we are in /home/name/mesh
 * the program will not cd inside mesh.
 */


void flush_output(FILE *fp, fltk::TextDisplay *shell_output) {
    char buffer[256];
    memset(buffer, 0, 256);
    while(fgets(buffer, 256, fp)) {
        printf("%s", buffer);
        shell_output->append(buffer);
    }
    pclose(fp);
    exit(0);
}
/********************************************************/


